from typing import Optional, Tuple

class DPMatrix:
    m: int
    n: int
    def __init__(self, reference: str, query: str): ...
    def set_entry(self, i: int, j: int, cost: int) -> None: ...
    def __str__(self) -> str: ...

class Aligner:
    def __init__(
        self,
        reference: str,
        max_error_rate: float,
        flags: int = 15,
        wildcard_ref: bool = False,
        wildcard_query: bool = False,
        indel_cost: int = 1,
        min_overlap: int = 1,
    ): ...
    def __repr__(self) -> str: ...
    def _set_reference(self, reference: str) -> None: ...
    @property
    def dpmatrix(self) -> DPMatrix: ...
    def enable_debug(self) -> None: ...
    def locate(self, query: str) -> Optional[Tuple[int, int, int, int, int, int]]: ...

class PrefixComparer:
    @property
    def effective_length(self) -> int: ...
    def __init__(
        self,
        reference: str,
        max_error_rate: float,
        wildcard_ref: bool = False,
        wildcard_query: bool = False,
        min_overlap: int = 1,
    ): ...
    def __repr__(self) -> str: ...
    def locate(self, query: str) -> Optional[Tuple[int, int, int, int, int, int]]: ...

class SuffixComparer(PrefixComparer): ...



"""
This module implements all the read modifications that cutadapt supports.
A modifier must be callable and typically implemented as a class with a
__call__ method.
"""
import re
import logging
from types import SimpleNamespace
from typing import Sequence, List, Tuple, Optional, Set
from abc import ABC, abstractmethod

from dnaio import record_names_match, SequenceRecord

from .qualtrim import quality_trim_index, nextseq_trim_index
from .adapters import (
    MultipleAdapters,
    SingleAdapter,
    IndexedPrefixAdapters,
    IndexedSuffixAdapters,
    Match,
    remainder,
    Adapter,
)
from .tokenizer import tokenize_braces, TokenizeError, Token, BraceToken
from .info import ModificationInfo

logger = logging.getLogger()


# If the number of prefix or suffix adapters is higher than this, switch to using an index
INDEXING_THRESHOLD = 5


class SingleEndModifier(ABC):
    @abstractmethod
    def __call__(self, read: SequenceRecord, info: ModificationInfo):
        pass


class PairedEndModifier(ABC):
    @abstractmethod
    def __call__(
        self,
        read1: SequenceRecord,
        read2: SequenceRecord,
        info1: ModificationInfo,
        info2: ModificationInfo,
    ) -> Tuple[SequenceRecord, SequenceRecord]:
        pass


class PairedEndModifierWrapper(PairedEndModifier):
    """
    Wrap two SingleEndModifiers that work on both reads in a paired-end read
    """

    paired = True

    def __init__(
        self,
        modifier1: Optional[SingleEndModifier],
        modifier2: Optional[SingleEndModifier],
    ):
        """Set one of the modifiers to None to work on R1 or R2 only"""
        self._modifier1 = modifier1
        self._modifier2 = modifier2
        if self._modifier1 is None and self._modifier2 is None:
            raise ValueError("Not both modifiers may be None")

    def __repr__(self):
        return (
            "PairedEndModifierWrapper("
            f"modifier1={self._modifier1!r}, modifier2={self._modifier2!r})"
        )

    def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
        if self._modifier1 is None:
            return read1, self._modifier2(read2, info2)  # type: ignore
        if self._modifier2 is None:
            return self._modifier1(read1, info1), read2
        return self._modifier1(read1, info1), self._modifier2(read2, info2)


class AdapterCutter(SingleEndModifier):
    """
    Repeatedly find one of multiple adapters in reads.

    Arguments:
        adapters: Adapters to be searched
        times: Repeat the search this number of times.
        action: What to do with a found adapter.
            - *None*: Do nothing, only update the ModificationInfo appropriately
            - "trim": Remove the adapter and down- or upstream sequence depending on adapter type
            - "mask": Replace the part of the sequence that would have been removed with "N" bases
            - "lowercase": Convert the part of the sequence that would have been removed to lowercase
            - "retain": Like "trim", but leave the adapter sequence itself in the read
        index: If True, attempt to create an index to speed up the search (if possible)
    """

    def __init__(
        self,
        adapters: Sequence[Adapter],
        times: int = 1,
        action: Optional[str] = "trim",
        index: bool = True,
    ):
        self.times = times
        assert action in ("trim", "mask", "lowercase", "retain", None)
        self.action = action
        self.with_adapters = 0
        self.adapter_statistics = {a: a.create_statistics() for a in adapters}
        if index:
            self.adapters = MultipleAdapters(
                self._regroup_into_indexed_adapters(adapters)
            )
        else:
            self.adapters = MultipleAdapters(adapters)
        if action == "retain" and times > 1:
            raise ValueError("'retain' cannot be combined with times > 1")
        if self.times == 1 and self.action == "trim":
            self.match_and_trim = self._match_and_trim_once_action_trim  # type: ignore

    def __repr__(self):
        return (
            "AdapterCutter("
            f"adapters={self.adapters!r}, times={self.times}, action='{self.action}')"
        )

    def _regroup_into_indexed_adapters(self, adapters):
        prefix, suffix, single = self._split_adapters(adapters)
        # For somewhat better backwards compatibility, avoid re-ordering
        # the adapters when we don’t need to
        if len(prefix) > INDEXING_THRESHOLD or len(suffix) > INDEXING_THRESHOLD:
            result = single
            if len(prefix) > 1:
                result.append(IndexedPrefixAdapters(prefix))
            else:
                result.extend(prefix)
            if len(suffix) > 1:
                result.append(IndexedSuffixAdapters(suffix))
            else:
                result.extend(suffix)
            return result
        else:
            return adapters

    @staticmethod
    def _split_adapters(
        adapters: Sequence[SingleAdapter],
    ) -> Tuple[
        Sequence[SingleAdapter], Sequence[SingleAdapter], Sequence[SingleAdapter]
    ]:
        """
        Split adapters into three different categories so that they can possibly be used
        with a MultiAdapter. Return a tuple (prefix, suffix, other), where
        - prefix is a list of all anchored 5' adapters that MultiAdapter would accept
        - suffix is a list of all anchored 3' adapters that MultiAdapter would accept
        - other is a list of all remaining adapters.
        """
        prefix: List[SingleAdapter] = []
        suffix: List[SingleAdapter] = []
        other: List[SingleAdapter] = []
        for a in adapters:
            if IndexedPrefixAdapters.is_acceptable(a):
                prefix.append(a)
            elif IndexedSuffixAdapters.is_acceptable(a):
                suffix.append(a)
            else:
                other.append(a)
        return prefix, suffix, other

    @staticmethod
    def trim_but_retain_adapter(read, matches: Sequence[Match]):
        start, stop = matches[-1].retained_adapter_interval()
        return read[start:stop]

    @staticmethod
    def masked_read(read, matches: Sequence[Match]):
        start, stop = remainder(matches)
        result = read[:]
        result.sequence = (
            "N" * start + read.sequence[start:stop] + "N" * (len(read) - stop)
        )
        return result

    @staticmethod
    def lowercased_read(read, matches: Sequence[Match]):
        start, stop = remainder(matches)
        result = read[:]
        result.sequence = (
            read.sequence[:start].lower()
            + read.sequence[start:stop].upper()
            + read.sequence[stop:].lower()
        )
        return result

    def __call__(self, read, info: ModificationInfo):
        trimmed_read, matches = self.match_and_trim(read)
        if matches:
            self.with_adapters += 1
            for match in matches:
                self.adapter_statistics[match.adapter].add_match(match)
        info.matches.extend(matches)  # TODO extend or overwrite?
        return trimmed_read

    def match_and_trim(self, read):
        """
        Search for the best-matching adapter in a read, perform the requested action
        ('trim', 'mask' etc. as determined by self.action) and return the
        (possibly) modified read.

        *self.times* adapter removal rounds are done. During each round,
        only the best-matching adapter is trimmed. If no adapter was found in a round,
        no further rounds are attempted.

        Return a pair (trimmed_read, matches), where matches is a list of Match instances.
        """
        matches = []
        if self.action == "lowercase":  # TODO this should not be needed
            read.sequence = read.sequence.upper()
        trimmed_read = read
        for _ in range(self.times):
            match = self.adapters.match_to(trimmed_read.sequence)
            if match is None:
                # if nothing found, attempt no further rounds
                break
            matches.append(match)
            trimmed_read = match.trimmed(trimmed_read)

        if not matches:
            return trimmed_read, []

        if self.action == "trim":
            # read is already trimmed, nothing to do
            pass
        elif self.action == "retain":
            trimmed_read = self.trim_but_retain_adapter(read, matches)
        elif self.action == "mask":
            trimmed_read = self.masked_read(read, matches)
        elif self.action == "lowercase":
            trimmed_read = self.lowercased_read(read, matches)
            assert len(trimmed_read.sequence) == len(read)
        elif self.action is None:
            trimmed_read = read[:]

        return trimmed_read, matches

    def _match_and_trim_once_action_trim(self, read):
        """
        Specalization of match_and_trim for the case that self.times == 1 and self.action == 'trim'
        """
        match = self.adapters.match_to(read.sequence)
        if match is not None:
            return match.trimmed(read), [match]
        else:
            return read, []


class ReverseComplementer(SingleEndModifier):
    """Trim adapters from a read and its reverse complement"""

    def __init__(self, adapter_cutter: AdapterCutter, rc_suffix: Optional[str] = " rc"):
        """
        rc_suffix -- suffix to add to the read name if sequence was reverse-complemented
        """
        self.adapter_cutter = adapter_cutter
        self.reverse_complemented = 0
        self._suffix = rc_suffix

    def __repr__(self):
        return f"ReverseComplementer(adapter_cutter={self.adapter_cutter})"

    def __call__(self, read: SequenceRecord, info: ModificationInfo):
        reverse_read = read.reverse_complement()

        forward_trimmed_read, forward_matches = self.adapter_cutter.match_and_trim(read)
        reverse_trimmed_read, reverse_matches = self.adapter_cutter.match_and_trim(
            reverse_read
        )

        forward_score = sum(m.score for m in forward_matches)
        reverse_score = sum(m.score for m in reverse_matches)
        use_reverse_complement = reverse_score > forward_score

        if use_reverse_complement:
            self.reverse_complemented += 1
            assert reverse_matches
            trimmed_read, matches = reverse_trimmed_read, reverse_matches
            info.is_rc = True
            if self._suffix:
                trimmed_read.name += self._suffix
        else:
            info.is_rc = False
            trimmed_read, matches = forward_trimmed_read, forward_matches

        if matches:
            self.adapter_cutter.with_adapters += 1
            for match in matches:
                stats = self.adapter_cutter.adapter_statistics[match.adapter]
                stats.add_match(match)
                stats.reverse_complemented += bool(use_reverse_complement)
            info.matches.extend(matches)  # TODO extend or overwrite?
        return trimmed_read


class PairedAdapterCutterError(Exception):
    pass


class PairedAdapterCutter(PairedEndModifier):
    """
    Trim adapters in pairs from R1 and R2.
    """

    def __init__(self, adapters1, adapters2, action="trim"):
        """
        adapters1 -- list of Adapters to be removed from R1
        adapters2 -- list of Adapters to be removed from R2

        Both lists must have the same, non-zero length.
         read pair is trimmed if adapters1[i] is found in R1 and adapters2[i] in R2.

        action -- What to do with a found adapter: None, 'trim', 'lowercase' or 'mask'
        """
        super().__init__()
        if len(adapters1) != len(adapters2):
            raise PairedAdapterCutterError(
                "The number of adapters to trim from R1 and R2 must be the same. "
                "Given: {} for R1, {} for R2".format(len(adapters1), len(adapters2))
            )
        if not adapters1:
            raise PairedAdapterCutterError("No adapters given")
        self._adapter_pairs = list(zip(adapters1, adapters2))
        logger.debug("Adapter pairs:")
        for a1, a2 in self._adapter_pairs:
            logger.debug(" • %s=%s -- %s=%s", a1.name, a1.spec(), a2.name, a2.spec())
        self.action = action
        self.with_adapters = 0
        self.adapter_statistics = [None, None]
        self.adapter_statistics[0] = {a: a.create_statistics() for a in adapters1}
        self.adapter_statistics[1] = {a: a.create_statistics() for a in adapters2}

    def __repr__(self):
        return f"PairedAdapterCutter(adapter_pairs={self._adapter_pairs!r})"

    def __call__(self, read1, read2, info1, info2):
        """ """
        best_matches = self._find_best_match_pair(read1.sequence, read2.sequence)
        if best_matches is None:
            return read1, read2
        match1, match2 = best_matches
        self.with_adapters += 1
        result = []
        for i, match, read in zip([0, 1], [match1, match2], [read1, read2]):
            trimmed_read = read
            if self.action == "lowercase":
                trimmed_read.sequence = trimmed_read.sequence.upper()

            trimmed_read = match.trimmed(trimmed_read)
            self.adapter_statistics[i][match.adapter].add_match(match)

            if self.action == "trim":
                # read is already trimmed, nothing to do
                pass
            elif self.action == "mask":
                trimmed_read = AdapterCutter.masked_read(read, [match])
            elif self.action == "lowercase":
                trimmed_read = AdapterCutter.lowercased_read(read, [match])
                assert len(trimmed_read.sequence) == len(read)
            elif self.action == "retain":
                trimmed_read = AdapterCutter.trim_but_retain_adapter(read, [match])
            elif self.action is None:  # --no-trim
                trimmed_read = read[:]
            result.append(trimmed_read)
        info1.matches.append(match1)
        info2.matches.append(match2)
        return result

    def _find_best_match_pair(
        self, sequence1: str, sequence2: str
    ) -> Optional[Tuple[Match, Match]]:
        best = None
        best_score = None
        best_errors = None
        for adapter1, adapter2 in self._adapter_pairs:
            match1 = adapter1.match_to(sequence1)
            if match1 is None:
                continue
            match2 = adapter2.match_to(sequence2)
            if match2 is None:
                continue
            total_score = match1.score + match2.score
            total_errors = match1.errors + match2.errors
            if (
                best is None
                or total_score > best_score
                or (total_score == best_score and total_errors < best_errors)
            ):
                best = match1, match2
                best_score = total_score
                best_errors = total_errors
        return best


class UnconditionalCutter(SingleEndModifier):
    """
    A modifier that unconditionally removes the first n or the last n bases from a read.

    If the length is positive, the bases are removed from the beginning of the read.
    If the length is negative, the bases are removed from the end of the read.
    """

    def __init__(self, length: int):
        self.length = length

    def __repr__(self):
        return f"UnconditionalCutter(length={self.length})"

    def __call__(self, read, info: ModificationInfo):
        if self.length > 0:
            info.cut_prefix = read.sequence[: self.length]
            return read[self.length :]
        elif self.length < 0:
            info.cut_suffix = read.sequence[self.length :]
            return read[: self.length]


class LengthTagModifier(SingleEndModifier):
    """
    Replace "length=..." strings in read names.
    """

    def __init__(self, length_tag):
        self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
        self.length_tag = length_tag

    def __repr__(self):
        return f"LengthTagModifier(length_tag='{self.length_tag}')"

    def __call__(self, read, info: ModificationInfo):
        read = read[:]
        if read.name.find(self.length_tag) >= 0:
            read.name = self.regex.sub(
                self.length_tag + str(len(read.sequence)), read.name
            )
        return read


class SuffixRemover(SingleEndModifier):
    """
    Remove a given suffix from read names.
    """

    def __init__(self, suffix):
        self.suffix = suffix

    def __repr__(self):
        return f"SuffixRemover('{self.suffix}')"

    def __call__(self, read, info: ModificationInfo):
        read = read[:]
        if read.name.endswith(self.suffix):
            read.name = read.name[: -len(self.suffix)]
        return read


class PrefixSuffixAdder(SingleEndModifier):
    """
    Add a suffix and a prefix to read names
    """

    def __init__(self, prefix, suffix):
        self.prefix = prefix
        self.suffix = suffix

    def __repr__(self):
        return f"PrefixSuffixAdder(prefix='{self.prefix}', suffix='{self.suffix}')"

    def __call__(self, read, info):
        read = read[:]
        adapter_name = info.matches[-1].adapter.name if info.matches else "no_adapter"
        read.name = (
            self.prefix.replace("{name}", adapter_name)
            + read.name
            + self.suffix.replace("{name}", adapter_name)
        )
        return read


class InvalidTemplate(Exception):
    pass


class Renamer(SingleEndModifier):
    """
    Rename reads using a template

    The template string can contain the following placeholders:

    - {header} -- full, unchanged header
    - {id} -- the part of the header before the first whitespace
    - {comment} -- the part of the header after the ID, excluding initial whitespace
    - {cut_prefix} -- prefix removed by UnconditionalCutter (with positive length argument)
    - {cut_suffix} -- suffix removed by UnconditionalCutter (with negative length argument)
    - {adapter_name} -- name of the *last* adapter match or no_adapter if there was none
    - {match_sequence} -- the sequence that matched the adapter (this includes possible errors)
          or an empty string if there was no match
    - {rc} -- the string 'rc' if the read was reverse complemented (with --revcomp) or '' otherwise
    """

    variables = {
        "header",
        "id",
        "comment",
        "cut_prefix",
        "cut_suffix",
        "adapter_name",
        "rc",
        "match_sequence",
    }

    def __init__(self, template: str):
        try:
            self._tokens = list(tokenize_braces(template))
        except TokenizeError as e:
            raise InvalidTemplate(f"Error in template '{template}': {e}")
        self.raise_if_invalid_variable(self._tokens, self.variables)
        self._template = template
        self._rename = self.compile_rename_function()

    def __repr__(self):
        return f"{self.__class__.__name__}('{self._template}')"

    def compile_rename_function(self):
        """
        Create the function that computes a new name

        By creating the code dynamically, we can ensure that only those placeholder values are
        computed that are actually used in the template.
        """
        code = {
            "header": "read.name",
            "id": "id_",
            "comment": "comment",
            "cut_prefix": "info.cut_prefix if info.cut_prefix else ''",
            "cut_suffix": "info.cut_suffix if info.cut_suffix else ''",
            "adapter_name": "info.matches[-1].adapter.name if info.matches else 'no_adapter'",
            "rc": "'rc' if info.is_rc else ''",
            "match_sequence": "info.matches[-1].match_sequence() if info.matches else ''",
        }
        placeholders = set(
            token.value for token in self._tokens if isinstance(token, BraceToken)
        )
        lines = ["def rename(self, read, info):"]
        if "id" in placeholders or "header" in placeholders:
            lines.append("  id_, comment = self.parse_name(read.name)")
        lines.append("  return self._template.format(")
        for placeholder in placeholders:
            lines.append(f"    {placeholder}={code[placeholder]},")
        lines.append("  )")
        logger.debug("Generated code of rename function:\n%s", "\n".join(lines))
        namespace = dict()
        exec("\n".join(lines), namespace)
        return namespace["rename"]

    @staticmethod
    def raise_if_invalid_variable(tokens: List[Token], allowed: Set[str]) -> None:
        for token in tokens:
            if not isinstance(token, BraceToken):
                continue
            value = token.value
            if value not in allowed:
                raise InvalidTemplate(
                    f"Error in template: Variable '{value}' not recognized"
                )

    @staticmethod
    def parse_name(read_name: str) -> Tuple[str, str]:
        """Parse read header and return (id, comment) tuple"""
        fields = read_name.split(maxsplit=1)
        if len(fields) == 2:
            return (fields[0], fields[1])
        else:
            return (read_name, "")

    def __call__(self, read: SequenceRecord, info: ModificationInfo) -> SequenceRecord:
        read.name = self._rename(self, read, info)
        return read


class PairedEndRenamer(PairedEndModifier):
    """
    Rename paired-end reads using a template. The template is applied to both
    R1 and R2, and the same template variables as in the (single-end) renamer
    are allowed. However,
    these variables are evaluated separately for each read. For example, if `{comment}`
    is used, it gets replaced with the R1 comment in the R1 header, and with the R2
    comment in the R2 header.

    Additionally, all template variables except `id` can be used in the read-specific
    forms `{r1.variablename}` and `{r2.variablename}`. For example, `{r1.comment}`
    always gets replaced with the R1 comment, even in R2.
    """

    def __init__(self, template: str):
        try:
            self._tokens = list(tokenize_braces(template))
        except TokenizeError as e:
            raise InvalidTemplate(f"Error in template '{template}': {e}")
        Renamer.raise_if_invalid_variable(self._tokens, self._get_allowed_variables())
        self._template = template

    @staticmethod
    def _get_allowed_variables() -> Set[str]:
        allowed = (Renamer.variables - {"rc"}) | {"rn"}
        for v in Renamer.variables - {"id", "rc"}:
            allowed.add("r1." + v)
            allowed.add("r2." + v)
        return allowed

    def __call__(
        self,
        read1: SequenceRecord,
        read2: SequenceRecord,
        info1: ModificationInfo,
        info2: ModificationInfo,
    ) -> Tuple[SequenceRecord, SequenceRecord]:

        if not record_names_match(read1.name, read2.name):
            id1 = Renamer.parse_name(read1.name)[0]
            id2 = Renamer.parse_name(read1.name)[1]
            raise ValueError(f"Input read IDs not identical: '{id1}' != '{id2}'")

        name1, name2 = self._rename(read1, read2, info1, info2)

        if not record_names_match(name1, name2):
            new_id1 = Renamer.parse_name(name1)[0]
            new_id2 = Renamer.parse_name(name2)[0]
            id1 = Renamer.parse_name(read1.name)[0]
            raise InvalidTemplate(
                "After renaming R1 and R2, their IDs are no longer identical: "
                f"'{new_id1}' != '{new_id2}'. Original read ID: '{id1}'. "
            )
        read1.name = name1
        read2.name = name2
        return read1, read2

    def _rename(
        self,
        read1: SequenceRecord,
        read2: SequenceRecord,
        info1: ModificationInfo,
        info2: ModificationInfo,
    ) -> Tuple[str, str]:

        id1, comment1 = Renamer.parse_name(read1.name)
        id2, comment2 = Renamer.parse_name(read2.name)
        header1 = read1.name
        header2 = read2.name

        d = []
        for id_, comment, header, info in (
            (id1, comment1, header1, info1),
            (id2, comment2, header2, info2),
        ):
            if info.matches:
                adapter_name = info.matches[-1].adapter.name
                match_sequence = info.matches[-1].match_sequence()
            else:
                adapter_name = "no_adapter"
                match_sequence = ""
            d.append(
                dict(
                    comment=comment,
                    header=header,
                    cut_prefix=info.cut_prefix if info.cut_prefix else "",
                    cut_suffix=info.cut_suffix if info.cut_suffix else "",
                    adapter_name=adapter_name,
                    match_sequence=match_sequence,
                )
            )
        name1 = self._template.format(
            id=id1,
            rn=1,
            **d[0],
            r1=SimpleNamespace(**d[0]),
            r2=SimpleNamespace(**d[1]),
        )
        name2 = self._template.format(
            id=id2,
            rn=2,
            **d[1],
            r1=SimpleNamespace(**d[0]),
            r2=SimpleNamespace(**d[1]),
        )
        return name1, name2


class ZeroCapper(SingleEndModifier):
    """
    Change negative quality values of a read to zero
    """

    def __init__(self, quality_base=33):
        self.quality_base = quality_base
        qb = quality_base
        self.zero_cap_trans = str.maketrans("".join(map(chr, range(qb))), chr(qb) * qb)

    def __repr__(self):
        return f"ZeroCapper(quality_base={self.quality_base})"

    def __call__(self, read, info: ModificationInfo):
        read = read[:]
        read.qualities = read.qualities.translate(self.zero_cap_trans)
        return read


class NextseqQualityTrimmer(SingleEndModifier):
    def __init__(self, cutoff: int, base: int = 33):
        self.cutoff = cutoff
        self.base = base
        self.trimmed_bases = 0

    def __repr__(self):
        return f"NextseqQualityTrimmer(cutoff={self.cutoff}, base={self.base})"

    def __call__(self, read, info: ModificationInfo):
        stop = nextseq_trim_index(read, self.cutoff, self.base)
        self.trimmed_bases += len(read) - stop
        return read[:stop]


class QualityTrimmer(SingleEndModifier):
    def __init__(self, cutoff_front: int, cutoff_back: int, base: int = 33):
        self.cutoff_front = cutoff_front
        self.cutoff_back = cutoff_back
        self.base = base
        self.trimmed_bases = 0

    def __repr__(self):
        return (
            f"QualityTrimmer(cutoff_front={self.cutoff_front}, "
            f"cutoff_back={self.cutoff_back}, base={self.base})"
        )

    def __call__(self, read, info: ModificationInfo):
        start, stop = quality_trim_index(
            read.qualities, self.cutoff_front, self.cutoff_back, self.base
        )
        self.trimmed_bases += len(read) - (stop - start)
        return read[start:stop]


class Shortener(SingleEndModifier):
    """Unconditionally shorten a read to the given length

    If the length is positive, the bases are removed from the end of the read.
    If the length is negative, the bases are removed from the beginning of the read.
    """

    def __init__(self, length):
        self.length = length

    def __repr__(self):
        return f"Shortener(length={self.length})"

    def __call__(self, read, info: ModificationInfo):
        if self.length >= 0:
            return read[: self.length]
        else:
            return read[self.length :]


class NEndTrimmer(SingleEndModifier):
    """Trims Ns from the 3' and 5' end of reads"""

    def __init__(self):
        self.start_trim = re.compile(r"^N+")
        self.end_trim = re.compile(r"N+$")

    def __repr__(self):
        return "NEndTrimmer()"

    def __call__(self, read, info: ModificationInfo):
        sequence = read.sequence
        start_cut = self.start_trim.match(sequence)
        end_cut = self.end_trim.search(sequence)
        start_cut = start_cut.end() if start_cut else 0
        end_cut = end_cut.start() if end_cut else len(read)
        return read[start_cut:end_cut]



import sys
import os.path
import subprocess


def datapath(path):
    return os.path.join(os.path.dirname(__file__), "data", path)


def cutpath(path):
    return os.path.join(os.path.dirname(__file__), "cut", path)


class FilesDifferent(Exception):
    pass


def assert_files_equal(path1, path2, ignore_trailing_space: bool = False):
    cmd = ["diff", "-u"]
    if sys.platform == "win32":
        cmd.append("--strip-trailing-cr")
    if ignore_trailing_space:
        if sys.platform == "darwin":
            # Ignores too much, but macOS doesn’t have the option below
            cmd.append("-b")
        else:
            cmd.append("--ignore-trailing-space")
    if sys.platform == "win32":
        path1, path2 = os.fspath(path1), os.fspath(path2)
    try:
        subprocess.check_output(cmd + [path1, path2], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        raise FilesDifferent("\n" + e.output.decode()) from None


def binomial(n, k):
    """
    Return binomial coefficient ('n choose k').
    This implementation does not use factorials.
    """
    k = min(k, n - k)
    if k < 0:
        return 0
    r = 1
    for j in range(k):
        r *= n - j
        r //= j + 1
    return r



import pytest

from cutadapt.tokenizer import tokenize_braces, StringToken, BraceToken, TokenizeError


def test_tokenize_braces():
    tokenize = tokenize_braces
    assert list(tokenize("")) == []
    assert list(tokenize("text")) == [StringToken("text")]
    assert list(tokenize("before {variable} after")) == [
        StringToken("before "),
        BraceToken("variable"),
        StringToken(" after"),
    ]


def test_tokenize_unexpected_braces():
    with pytest.raises(TokenizeError):
        list(tokenize_braces("abc {def{ghi}"))

    with pytest.raises(TokenizeError):
        list(tokenize_braces("abc {def} gh} i"))



from collections import defaultdict, Counter
from typing import DefaultDict, Tuple


class ReadLengthStatistics:
    """
    Keep track of the lengths of written reads or read pairs
    """

    def __init__(self) -> None:
        # It would be more natural to use a Counter, but a
        # defaultdict is much faster
        self._written_lengths1: DefaultDict[int, int] = defaultdict(int)
        self._written_lengths2: DefaultDict[int, int] = defaultdict(int)

    def update(self, read) -> None:
        """Add a single-end read to the statistics"""
        self._written_lengths1[len(read)] += 1

    def update2(self, read1, read2) -> None:
        """Add a paired-end read to the statistics"""
        self._written_lengths1[len(read1)] += 1
        self._written_lengths2[len(read2)] += 1

    def written_reads(self) -> int:
        """Return number of written reads or read pairs"""
        return sum(self._written_lengths1.values())

    def written_bp(self) -> Tuple[int, int]:
        return (
            self._compute_total_bp(self._written_lengths1),
            self._compute_total_bp(self._written_lengths2),
        )

    def written_lengths(self) -> Tuple[Counter, Counter]:
        return (Counter(self._written_lengths1), Counter(self._written_lengths2))

    @staticmethod
    def _compute_total_bp(counts: DefaultDict[int, int]) -> int:
        return sum(length * count for length, count in counts.items())

    def __iadd__(self, other):
        written_lengths1, written_lengths2 = other.written_lengths()
        for length, count in written_lengths1.items():
            self._written_lengths1[length] += count
        for length, count in written_lengths2.items():
            self._written_lengths2[length] += count
        return self



from cutadapt.report import (
    safe_divide,
    add_if_not_none,
)


def test_safe_divide():
    assert safe_divide(1, 0) == 0
    assert safe_divide(5, 2) == 2.5


def test_add_if_not_none():
    assert add_if_not_none(3, 5) == 8
    assert add_if_not_none(3, None) == 3
    assert add_if_not_none(None, 5) == 5



import os

import pytest
from utils import assert_files_equal, datapath, cutpath
from cutadapt.__main__ import main
from cutadapt.report import Statistics


@pytest.fixture(params=[1, 2])
def cores(request):
    return request.param


@pytest.fixture
def run(tmp_path):
    def _run(params, expected, inpath) -> Statistics:
        if type(params) is str:
            params = params.split()
        params += ["--json", os.fspath(tmp_path / "stats.cutadapt.json")]
        tmp_fastaq = tmp_path / expected
        params += ["-o", tmp_fastaq]
        params += [datapath(inpath)]
        stats = main([str(p) for p in params])
        # TODO redirect standard output
        assert_files_equal(cutpath(expected), tmp_fastaq)
        return stats

    return _run



"""
Parse adapter specifications
"""
import re
import logging
from pathlib import Path
from typing import Type, Optional, List, Tuple, Any, Dict, Iterable
from xopen import xopen
from dnaio.readers import FastaReader
from .adapters import (
    Adapter,
    FrontAdapter,
    NonInternalFrontAdapter,
    BackAdapter,
    NonInternalBackAdapter,
    AnywhereAdapter,
    PrefixAdapter,
    SuffixAdapter,
    LinkedAdapter,
    InvalidCharacter,
    RightmostFrontAdapter,
)

logger = logging.getLogger(__name__)


def parse_search_parameters(spec: str):
    """Parse key=value;key=value;key=value into a dict"""

    allowed_parameters = {
        # abbreviations
        "e": "max_error_rate",
        "error_rate": "max_errors",
        "max_error_rate": "max_errors",
        "o": "min_overlap",
        # allowed parameters
        "max_errors": None,
        "min_overlap": None,
        "anywhere": None,
        "required": None,
        "optional": None,  # If this is specified, 'required' will be set to False
        "indels": None,
        "noindels": None,
        "rightmost": None,
    }

    fields = spec.split(";")
    result: Dict[str, Any] = dict()
    for field in fields:
        field = field.strip()
        if not field:
            continue
        key, equals, value = field.partition("=")  # type: (str, str, Any)
        key = key.strip()
        if key not in allowed_parameters:
            raise KeyError(f"Unknown parameter '{key}'")
        if equals == "=" and value == "":
            raise ValueError(f"No value given for key '{key}'")
        # unabbreviate
        while allowed_parameters[key] is not None:
            key = allowed_parameters[key]  # type: ignore
        value = value.strip()
        if value == "":
            value = True
        else:
            try:
                value = int(value)
            except ValueError:
                value = float(value)
        if key in result:
            raise KeyError(f"Key '{key}' specified twice")
        result[key] = value
    if "optional" in result and "required" in result:
        raise ValueError(
            "'optional' and 'required' cannot be specified at the same time"
        )
    if "indels" in result and "noindels" in result:
        raise ValueError("'indels' and 'noindels' cannot be specified at the same time")
    if "optional" in result:
        result["required"] = False
        del result["optional"]
    if "noindels" in result:
        result["indels"] = False
        del result["noindels"]
    return result


def expand_braces(sequence: str) -> str:
    """
    Replace all occurrences of ``x{n}`` (where x is any character) with n
    occurrences of x. Raise ValueError if the expression cannot be parsed.

    >>> expand_braces('TGA{5}CT')
    'TGAAAAACT'
    """
    # Simple DFA with four states, encoded in prev
    result = ""
    prev = None
    for s in re.split("([{}])", sequence):
        if s == "":
            continue
        if prev is None:
            if s == "{":
                raise ValueError('"{" must be used after a character')
            if s == "}":
                raise ValueError('"}" cannot be used here')
            prev = s
            result += s
        elif prev == "{":
            prev = int(s)
            if not 0 <= prev <= 10000:
                raise ValueError(f"Value {prev} invalid")
        elif isinstance(prev, int):
            if s != "}":
                raise ValueError('"}" expected')
            result = result[:-1] + result[-1] * prev
            prev = None
        else:
            if s != "{":
                raise ValueError('Expected "{"')
            prev = "{"
    # Check if we are in a non-terminating state
    if isinstance(prev, int) or prev == "{":
        raise ValueError("Unterminated expression")
    return result


def _normalize_ellipsis(spec1: str, spec2: str, adapter_type) -> Tuple[str, str]:
    if adapter_type == "anywhere":
        raise ValueError('No ellipsis ("...") allowed in "anywhere" adapters')
    if not spec1:
        if adapter_type == "back":
            # -a ...ADAPTER
            spec = spec2
        else:
            # -g ...ADAPTER
            raise ValueError("Invalid adapter specification")
    elif not spec2:
        if adapter_type == "back":
            # -a ADAPTER...
            adapter_type = "front"
            spec = spec1
        else:
            # -g ADAPTER...
            spec = spec1
    else:
        raise ValueError("Expected either spec1 or spec2")
    return spec, adapter_type


class AdapterSpecification:
    """# noqa: E501
    Description of a single non-linked adapter.

    These are the attributes:

    - name (None or str)
    - restriction (None, 'anchored', or 'noninternal')
    - sequence (nucleotide sequence as string)
    - search parameters (dict with keys such as 'max_errors', 'min_overlap')
    - adapter_type ('front' for -a, 'back' for -g and 'anywhere' for -b)

    >>> AdapterSpecification.parse('a_name=ACGT;anywhere', 'back')
    AdapterSpecification(name='a_name', restriction=None, sequence='ACGT', parameters={'anywhere': True}, adapter_type='back')
    """

    def __init__(
        self,
        name: Optional[str],
        restriction: Optional[str],
        sequence: str,
        parameters,
        adapter_type: str,
        rightmost: bool,
    ):
        assert restriction in (None, "anchored", "noninternal")
        assert adapter_type in ("front", "back", "anywhere")
        self.name = name
        self.restriction = restriction
        self.sequence = sequence
        self.parameters = parameters
        self.adapter_type = adapter_type
        self.rightmost = rightmost

    def __repr__(self):
        return "{}(name={!r}, restriction={!r}, sequence={!r}, parameters={!r}, adapter_type={!r})".format(
            self.__class__.__name__,
            self.name,
            self.restriction,
            self.sequence,
            self.parameters,
            self.adapter_type,
        )

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.restriction == other.restriction
            and self.sequence == other.sequence
            and self.parameters == other.parameters
            and self.adapter_type == other.adapter_type
        )

    @staticmethod
    def _extract_name(spec: str) -> Tuple[Optional[str], str]:
        """
        Parse an adapter specification given as 'name=adapt' into 'name' and 'adapt'.
        """
        fields = spec.split("=", 1)
        name: Optional[str] = None
        if len(fields) > 1:
            name, spec = fields
            name = name.strip()
        spec = spec.strip()
        return name, spec

    @classmethod
    def parse(cls, spec: str, adapter_type: str) -> "AdapterSpecification":
        """
        Parse an adapter specification for a non-linked adapter (without '...')
        and return an AdapterSpecification instance.

        Allow:
        'back' and ADAPTER
        'back' and ADAPTERX
        'back' and ADAPTER$
        'front' and ADAPTER
        'front' and XADAPTER
        'front' and ^ADAPTER
        'anywhere' and ADAPTER
        """
        if adapter_type not in ("front", "back", "anywhere"):
            raise ValueError("adapter_type must be front, back or anywhere")

        spec, middle, parameters_spec = spec.partition(";")
        name, spec = cls._extract_name(spec)
        spec = spec.strip()
        parameters = parse_search_parameters(parameters_spec)
        spec = expand_braces(spec)
        rightmost = parameters.pop("rightmost", False)

        # Special case for adapters consisting of only X characters:
        # This needs to be supported for backwards-compatibilitity
        if len(spec.strip("X")) == 0:
            return cls(name, None, spec, {}, adapter_type, False)

        try:
            front_restriction, back_restriction, spec = cls._parse_restrictions(spec)
        except ValueError:
            raise ValueError(
                "You cannot use multiple placement restrictions for an adapter at the same time. "
                "Choose one of ^ADAPTER, ADAPTER$, XADAPTER or ADAPTERX"
            ) from None

        if adapter_type == "front" and back_restriction:
            raise ValueError(
                "Allowed placement restrictions for a 5' adapter are XADAPTER and ^ADAPTER"
            )
        if adapter_type == "back" and front_restriction:
            raise ValueError(
                "Allowed placement restrictions for a 3' adapter are ADAPTERX and ADAPTER$"
            )

        if front_restriction is not None:
            restriction: Optional[str] = front_restriction
        else:
            restriction = back_restriction

        if adapter_type == "anywhere" and restriction is not None:
            raise ValueError(
                "Placement restrictions (with X, ^, $) not supported for 'anywhere' (-b) adapters"
            )

        if "min_overlap" in parameters and restriction == "anchored":
            raise ValueError(
                "Setting 'min_overlap=' (or 'o=') for anchored adapters is not possible because "
                "anchored adapters always need to match in full."
            )

        if parameters.get("min_overlap", 0) > len(spec):
            raise ValueError(
                f"min_overlap={parameters['min_overlap']}"
                f" exceeds length of adapter {spec}"
            )

        if rightmost and (adapter_type != "front" or restriction is not None):
            raise ValueError("'rightmost' only allowed with regular 5' adapters")

        return cls(name, restriction, spec, parameters, adapter_type, rightmost)

    @staticmethod
    def _parse_restrictions(spec: str) -> Tuple[Optional[str], Optional[str], str]:
        front_restriction = None
        if spec.startswith("^"):
            front_restriction = "anchored"
            spec = spec[1:]
        if spec.upper().startswith("X"):
            if front_restriction is not None:
                raise ValueError("two front restrictions")
            front_restriction = "noninternal"
            spec = spec.lstrip("xX")

        back_restriction = None
        if spec.endswith("$"):
            back_restriction = "anchored"
            spec = spec[:-1]
        if spec.upper().endswith("X"):
            if back_restriction is not None:
                raise ValueError("two back restrictions")
            back_restriction = "noninternal"
            spec = spec.rstrip("xX")

        n_placement_restrictions = int(bool(front_restriction)) + int(
            bool(back_restriction)
        )
        if n_placement_restrictions > 1:
            raise ValueError("front and back restrictions")
        assert front_restriction is None or back_restriction is None
        return front_restriction, back_restriction, spec

    @staticmethod
    def _restriction_to_class(adapter_type, restriction, rightmost):
        """
        restriction: None, "anchored", or "noninternal"
        """
        if adapter_type == "front":
            if rightmost:
                assert restriction is None
                return RightmostFrontAdapter
            elif restriction is None:
                return FrontAdapter
            elif restriction == "anchored":
                return PrefixAdapter
            elif restriction == "noninternal":
                return NonInternalFrontAdapter
            else:
                raise ValueError(
                    f"Value {restriction} for a front restriction not allowed"
                )
        elif adapter_type == "back":
            if restriction is None:
                return BackAdapter
            elif restriction == "anchored":
                return SuffixAdapter
            elif restriction == "noninternal":
                return NonInternalBackAdapter
            else:
                raise ValueError(
                    f"Value {restriction} for a back restriction not allowed"
                )
        else:
            assert adapter_type == "anywhere"
            if restriction is None:
                return AnywhereAdapter
            else:
                raise ValueError(
                    'No placement may be specified for "anywhere" adapters'
                )

    def adapter_class(self):
        return self._restriction_to_class(
            self.adapter_type, self.restriction, self.rightmost
        )


def make_adapters_from_specifications(
    type_spec_pairs: List[Tuple[str, str]],
    search_parameters: Dict[str, Any],
) -> List[Adapter]:
    """
    Create a list of Adapter classes from specification strings and adapter types.

    type_spec_pairs -- a list of (str, str) pairs, where the first is
      the adapter type (either 'front', 'back' or 'anywhere') and the second is the
      adapter specification string, such as "ACGT;o=3" or "file:adapters.fasta"

    search_parameters -- A dict with default search parameters. These can be overriden by the
      adapter specifications. They are passed as **kwargs when instantiating the
      adapter classes.
      Possible keys: max_error_rate, min_overlap, read_wildcards, adapter_wildcards, indels

    Return a list of appropriate Adapter instances.
    """
    adapters: List[Adapter] = []
    for adapter_type, spec in type_spec_pairs:
        adapters.extend(
            make_adapters_from_one_specification(spec, adapter_type, search_parameters)
        )
    return adapters


def make_adapters_from_one_specification(
    spec: str,
    adapter_type: str,
    search_parameters: Dict[str, Any],
) -> Iterable[Adapter]:
    """
    Parse an adapter specification and yield appropriate Adapter classes.
    """
    if spec.startswith("file:") or spec.startswith("^file:"):
        if spec.startswith("^"):
            spec = spec[1:]
            anchoring_prefix = "^"
        else:
            anchoring_prefix = ""
        path, _, parameters_spec = spec[5:].partition(";")
        parameters = search_parameters.copy()
        parameters.update(parse_search_parameters(parameters_spec))
        for name, spec in read_adapters_fasta(path):
            yield make_adapter(
                anchoring_prefix + spec, adapter_type, parameters, name=name
            )
    else:
        try:
            yield make_adapter(spec, adapter_type, search_parameters)
        except InvalidCharacter as e:
            if Path(spec).exists():
                extra_message = (
                    f"A file exists named '{spec}'. "
                    "To use the sequences in that file as adapter sequences, write 'file:' "
                    f"before the path, as in 'file:{spec}'."
                )
                raise InvalidCharacter(e.args[0] + "\n" + extra_message)
            else:
                raise


def make_adapter(
    spec: str,
    adapter_type: str,
    search_parameters: Dict[str, Any],
    name: Optional[str] = None,
) -> Adapter:
    """
    Parse an adapter specification not using ``file:`` notation and return
    an object of an appropriate Adapter class.

    name -- Adapter name if not included as part of the spec. (If spec is
    'name=ADAPTER', name will be 'name'.)

    adapter_type -- describes which commandline parameter was used (``-a``
    is 'back', ``-b`` is 'anywhere', and ``-g`` is 'front').

    search_parameters -- dict with default search parameters
    """
    if adapter_type not in ("front", "back", "anywhere"):
        raise ValueError("adapter_type must be front, back or anywhere")
    spec1, middle, spec2 = spec.partition("...")
    if middle == "..." and spec1 and spec2:
        return _make_linked_adapter(spec1, spec2, name, adapter_type, search_parameters)

    if middle == "...":
        spec, adapter_type = _normalize_ellipsis(spec1, spec2, adapter_type)
    else:
        spec = spec1
    return _make_not_linked_adapter(spec, name, adapter_type, search_parameters)


def _make_linked_adapter(
    spec1: str,
    spec2: str,
    name: Optional[str],
    adapter_type: str,
    search_parameters: Dict[str, Any],
) -> LinkedAdapter:
    """Return a linked adapter from two specification strings"""

    if adapter_type == "anywhere":
        raise ValueError("'anywhere' (-b) adapters may not be linked")
    front_spec = AdapterSpecification.parse(spec1, "front")
    back_spec = AdapterSpecification.parse(spec2, "back")
    if name is None:
        name = front_spec.name

    front_anchored = front_spec.restriction is not None
    back_anchored = back_spec.restriction is not None

    front_parameters = search_parameters.copy()
    front_parameters.update(front_spec.parameters)
    back_parameters = search_parameters.copy()
    back_parameters.update(back_spec.parameters)

    if adapter_type == "front":
        # -g requires both adapters to be present
        front_required = True
        back_required = True
    else:
        # -a requires only the anchored adapters to be present
        front_required = front_anchored
        back_required = back_anchored

    # Handle parameters overriding whether an adapter is required
    front_required = front_parameters.pop("required", front_required)
    back_required = back_parameters.pop("required", back_required)

    front_adapter = front_spec.adapter_class()(
        front_spec.sequence, name="linked_front", **front_parameters
    )
    back_adapter = back_spec.adapter_class()(
        back_spec.sequence, name="linked_back", **back_parameters
    )

    return LinkedAdapter(
        front_adapter=front_adapter,
        back_adapter=back_adapter,
        front_required=front_required,
        back_required=back_required,
        name=name,
    )


def _make_not_linked_adapter(
    spec: str,
    name: Optional[str],
    adapter_type: str,
    search_parameters: Dict[str, Any],
) -> Adapter:
    aspec = AdapterSpecification.parse(spec, adapter_type)
    adapter_class: Type[Adapter] = aspec.adapter_class()

    if aspec.parameters.pop("anywhere", False) and adapter_class in (
        FrontAdapter,
        BackAdapter,
        RightmostFrontAdapter,
    ):
        aspec.parameters["force_anywhere"] = True
    if "required" in aspec.parameters:
        raise ValueError(
            "'required' and 'optional' can only be used within linked adapters"
        )
    parameters = search_parameters.copy()
    parameters.update(aspec.parameters)
    return adapter_class(
        sequence=aspec.sequence,
        name=aspec.name if name is None else name,
        **parameters,
    )


def read_adapters_fasta(path):
    """
    Read adapter sequences from a FASTA file
    """
    with xopen(path, mode="rb", threads=0) as f:
        fasta = FastaReader(f)  # type: ignore
        for record in fasta:
            header = record.name.split(None, 1)
            name = header[0] if header else None
            yield name, record.sequence



# cython: profile=False, emit_code_comments=False, language_level=3
"""
Quality trimming.
"""

cdef extern from *:
    unsigned char * PyUnicode_1BYTE_DATA(object o)
    int PyUnicode_KIND(object o)
    int PyUnicode_1BYTE_KIND


cdef class HasNoQualities(Exception):
    pass


def quality_trim_index(str qualities, int cutoff_front, int cutoff_back, int base=33):
    """
    Find the positions at which to trim low-quality ends from a nucleotide sequence.
    Return tuple (start, stop) that indicates the good-quality segment.

    Qualities are assumed to be ASCII-encoded as chr(qual + base).

    The algorithm is the same as the one used by BWA within the function
    'bwa_trim_read':
    - Subtract the cutoff value from all qualities.
    - Compute partial sums from all indices to the end of the sequence.
    - Trim sequence at the index at which the sum is minimal.
    """
    if qualities is None:
        raise HasNoQualities("Cannot do quality trimming when no qualities are available")
    cdef:
        int s
        int max_qual
        int n = len(qualities)
        int stop = n
        int start = 0
        int i
        char* qual

    if not PyUnicode_KIND(qualities) == PyUnicode_1BYTE_KIND:
        raise ValueError("Quality data is not ASCII")
    qual = <char *>PyUnicode_1BYTE_DATA(qualities)

    # find trim position for 5' end
    s = 0
    max_qual = 0
    for i in range(n):
        s += cutoff_front - (qual[i] - base)
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            start = i + 1

    # same for 3' end
    max_qual = 0
    s = 0
    for i in reversed(range(n)):
        s += cutoff_back - (qual[i] - base)
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            stop = i
    if start >= stop:
        start, stop = 0, 0
    return (start, stop)


def nextseq_trim_index(sequence, int cutoff, int base=33):
    """
    Variant of the above quality trimming routine that works on NextSeq data.
    With Illumina NextSeq, bases are encoded with two colors. 'No color' (a
    dark cycle) usually means that a 'G' was sequenced, but that also occurs
    when sequencing falls off the end of the fragment. The read then contains
    a run of high-quality G bases in the end.

    This routine works as the one above, but counts qualities belonging to 'G'
    bases as being equal to cutoff - 1.
    """
    bases = sequence.sequence
    qualities = sequence.qualities
    if qualities is None:
        raise HasNoQualities()
    cdef:
        int s = 0
        int max_qual = 0
        int max_i
        int i, q
        char* qual

    if not PyUnicode_KIND(qualities) == PyUnicode_1BYTE_KIND:
        raise ValueError("Quality data is not ASCII")
    qual = <char *>PyUnicode_1BYTE_DATA(qualities)

    s = 0
    max_qual = 0
    max_i = len(qualities)
    for i in reversed(range(max_i)):
        q = qual[i] - base
        if bases[i] == 'G':
            q = cutoff - 1
        s += cutoff - q
        if s < 0:
            break
        if s > max_qual:
            max_qual = s
            max_i = i
    return max_i


def expected_errors(str qualities, int base=33):
    """
    Return the number of expected errors (as double) from a read’s
    qualities.

    This uses the formula in Edgar et al. (2015),
    see Section 2.2 in <https://academic.oup.com/bioinformatics/article/31/21/3476/194979>.

    qualities -- ASCII-encoded qualities (chr(qual + base))
    """
    cdef:
        int i, q
        bytes quals = qualities.encode()
        char* cq = quals
        double e = 0.0

    for i in range(len(qualities)):
        q = cq[i] - base
        e += 10 ** (-q / 10)
    return e



import time
from itertools import islice

from cutadapt.utils import (
    raise_open_files_limit,
    Progress,
    DummyProgress,
)


def test_raise_open_files_limit():
    try:
        raise_open_files_limit(1)
    except ValueError:
        pass


def test_progress():
    p = Progress(every=1e-6)
    p.update(100)
    time.sleep(0.001)
    p.update(0)
    p.update(900)
    p.update(10000)
    p.close()


def test_progress_scissors():
    sc = Progress.scissors(width=10)
    for i in islice(sc, 0, 30):
        next(sc)


def test_dummy_progress():
    p = DummyProgress()
    p.update(100)
    p.update(900)
    p.close()



import sys
import logging


# Custom log level
REPORT = 25


class CrashingHandler(logging.StreamHandler):
    def emit(self, record):
        """Unlike the method it overrides, this will not catch exceptions"""
        msg = self.format(record)
        stream = self.stream
        stream.write(msg)
        stream.write(self.terminator)
        self.flush()


class NiceFormatter(logging.Formatter):
    """
    Do not prefix "INFO:" to info-level log messages (but do it for all other
    levels).

    Based on http://stackoverflow.com/a/9218261/715090 .
    """

    def format(self, record):
        if record.levelno not in (logging.INFO, REPORT):
            record.msg = f"{record.levelname}: {record.msg}"
        return super().format(record)


def setup_logging(logger, log_to_stderr=True, minimal=False, quiet=False, debug=0):
    """
    Attach handler to the global logger object
    """
    # For --report=minimal, we need this custom log level because we want to
    # print nothing except the minimal report and therefore cannot use the
    # INFO level (and the ERROR level would give us an 'ERROR:' prefix).
    logging.addLevelName(REPORT, "REPORT")

    stream_handler = CrashingHandler(sys.stderr if log_to_stderr else sys.stdout)
    stream_handler.setFormatter(NiceFormatter())
    # debug overrides quiet overrides minimal
    if debug > 0:
        level = logging.DEBUG
    elif quiet:
        level = logging.ERROR
    elif minimal:
        level = REPORT
    else:
        level = logging.INFO
    stream_handler.setLevel(level)
    logger.setLevel(level)
    logger.addHandler(stream_handler)



"""
Adapter finding and trimming classes

The ...Adapter classes are responsible for finding adapters.
The ...Match classes trim the reads.
"""
import logging
from enum import IntFlag
from collections import defaultdict
from typing import Optional, Tuple, Sequence, Dict, Any, List, Union
from abc import ABC, abstractmethod

from .align import (
    EndSkip,
    Aligner,
    PrefixComparer,
    SuffixComparer,
    edit_environment,
    hamming_environment,
)

logger = logging.getLogger()


class InvalidCharacter(Exception):
    pass


# TODO remove this enum, this should be within each Adapter class
class Where(IntFlag):
    """
    Aligner flag combinations for all adapter types.

    "REFERENCE" is the adapter sequence, "QUERY" is the read sequence
    """

    BACK = EndSkip.QUERY_START | EndSkip.QUERY_STOP | EndSkip.REFERENCE_END
    FRONT = EndSkip.QUERY_START | EndSkip.QUERY_STOP | EndSkip.REFERENCE_START
    PREFIX = EndSkip.QUERY_STOP
    SUFFIX = EndSkip.QUERY_START
    # Just like FRONT/BACK, but without internal matches
    FRONT_NOT_INTERNAL = EndSkip.REFERENCE_START | EndSkip.QUERY_STOP
    BACK_NOT_INTERNAL = EndSkip.QUERY_START | EndSkip.REFERENCE_END
    ANYWHERE = EndSkip.SEMIGLOBAL


def returns_defaultdict_int():
    # We need this function to make EndStatistics picklable.
    # Even a @staticmethod of EndStatistics is not sufficient
    # as that is not picklable before Python 3.5.
    return defaultdict(int)


class EndStatistics:
    """Statistics about the 5' or 3' end"""

    def __init__(self, adapter: "SingleAdapter"):
        self.max_error_rate: float = adapter.max_error_rate
        self.sequence: str = adapter.sequence
        self.effective_length: int = adapter.effective_length
        self.has_wildcards: bool = adapter.adapter_wildcards
        self.indels: bool = adapter.indels
        self.adapter_type: str = adapter.descriptive_identifier()
        self.allows_partial_matches: bool = adapter.allows_partial_matches
        # self.errors[l][e] == n iff a sequence of length l matching at e errors was removed n times
        self.errors: Dict[int, Dict[int, int]] = defaultdict(returns_defaultdict_int)
        self.adjacent_bases = {"A": 0, "C": 0, "G": 0, "T": 0, "": 0}
        # TODO avoid hard-coding the list of classes
        self._remove_prefix = isinstance(adapter, FrontAdapter)

    def __repr__(self):
        errors = {k: dict(v) for k, v in self.errors.items()}
        return "EndStatistics(max_error_rate={}, errors={}, adjacent_bases={})".format(
            self.max_error_rate,
            errors,
            self.adjacent_bases,
        )

    def __iadd__(self, other: Any):
        if not isinstance(other, self.__class__):
            raise ValueError("Cannot compare")
        if (
            self.max_error_rate != other.max_error_rate
            or self.sequence != other.sequence
            or self.effective_length != other.effective_length
            or self.indels != other.indels
        ):
            raise RuntimeError("Incompatible EndStatistics, cannot be added")
        for base in ("A", "C", "G", "T", ""):
            self.adjacent_bases[base] += other.adjacent_bases[base]
        for length, error_dict in other.errors.items():
            for errors in error_dict:
                self.errors[length][errors] += other.errors[length][errors]
        return self

    @property
    def lengths(self):
        d = {length: sum(errors.values()) for length, errors in self.errors.items()}
        return d

    def random_match_probabilities(self, gc_content: float) -> List[float]:
        """
        Estimate probabilities that this adapter end matches a
        random sequence. Indels are not taken into account.

        Returns a list p, where p[i] is the probability that
        i bases of this adapter match a random sequence with
        GC content gc_content.
        """
        assert 0.0 <= gc_content <= 1.0
        seq = self.sequence
        # FIXME this is broken for 'anywhere' adapters
        if self._remove_prefix:
            seq = seq[::-1]
        allowed_bases = "CGRYSKMBDHVN" if self.has_wildcards else "GC"
        p = 1.0
        probabilities = [p]
        for i, c in enumerate(seq):
            if c in allowed_bases:
                p *= gc_content / 2.0
            else:
                p *= (1.0 - gc_content) / 2.0
            probabilities.append(p)
        return probabilities


class AdapterStatistics(ABC):

    reverse_complemented: int = 0
    name: str
    adapter: "Adapter"

    @abstractmethod
    def __iadd__(self, other):
        pass

    @abstractmethod
    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
        pass

    @abstractmethod
    def add_match(self, match) -> None:
        pass


class SingleAdapterStatistics(AdapterStatistics, ABC):
    """
    Statistics about a 5' or 3' adapter, where we only need to keep track of sequences
    removed from one "end".
    """

    def __init__(self, adapter: "SingleAdapter"):
        self.name = adapter.name
        self.adapter = adapter
        self.end = EndStatistics(adapter)

    def __repr__(self):
        return f"SingleAdapterStatistics(name={self.name}, end={self.end})"

    def __iadd__(self, other: "SingleAdapterStatistics"):
        if not isinstance(other, self.__class__):
            raise ValueError("Cannot iadd")
        self.end += other.end
        self.reverse_complemented += other.reverse_complemented
        return self


class FrontAdapterStatistics(SingleAdapterStatistics):
    def add_match(self, match: "RemoveBeforeMatch"):
        self.end.errors[match.removed_sequence_length()][match.errors] += 1

    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
        return self.end, None


class BackAdapterStatistics(SingleAdapterStatistics):
    def add_match(self, match: "RemoveAfterMatch"):
        adjacent_base = match.adjacent_base()
        self.end.errors[match.removed_sequence_length()][match.errors] += 1
        try:
            self.end.adjacent_bases[adjacent_base] += 1
        except KeyError:
            self.end.adjacent_bases[""] += 1

    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
        return None, self.end


class LinkedAdapterStatistics(AdapterStatistics):
    """
    Statistics about sequences removed by a lined adapter.
    """

    def __init__(
        self,
        adapter: "LinkedAdapter",
        front: "SingleAdapter",
        back: "SingleAdapter",
    ):
        self.name = adapter.name
        self.adapter = adapter
        self.front = EndStatistics(front)
        self.back = EndStatistics(back)
        self.reverse_complemented = 0

    def __repr__(self):
        return f"LinkedAdapterStatistics(name={self.name}, front={self.front}, back={self.back})"

    def __iadd__(self, other: "LinkedAdapterStatistics"):
        if not isinstance(other, self.__class__):
            raise ValueError("Cannot iadd")
        self.front += other.front
        self.back += other.back
        self.reverse_complemented += other.reverse_complemented
        return self

    def add_match(self, match: "LinkedMatch"):
        # TODO this is duplicated code
        if match.front_match:
            self.front.errors[match.front_match.removed_sequence_length()][
                match.front_match.errors
            ] += 1
        if match.back_match:
            adjacent_base = match.back_match.adjacent_base()
            self.back.errors[match.back_match.removed_sequence_length()][
                match.back_match.errors
            ] += 1
            try:
                self.back.adjacent_bases[adjacent_base] += 1
            except KeyError:
                self.back.adjacent_bases[""] += 1

    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
        return self.front, self.back


class AnywhereAdapterStatistics(AdapterStatistics):
    """
    Statistics about sequences removed by a lined adapter.
    """

    def __init__(self, adapter: "AnywhereAdapter"):
        self.name = adapter.name
        self.adapter = adapter
        self.front = EndStatistics(adapter)
        self.back = EndStatistics(adapter)
        self.reverse_complemented = 0

    def __repr__(self):
        return f"AnywhereAdapterStatistics(name={self.name}, front={self.front}, back={self.back})"

    def __iadd__(self, other: "AnywhereAdapterStatistics"):
        if not isinstance(other, AnywhereAdapterStatistics):
            raise ValueError("Cannot add")
        self.front += other.front
        self.back += other.back
        self.reverse_complemented += other.reverse_complemented
        return self

    def add_match(self, match: Union["RemoveBeforeMatch", "RemoveAfterMatch"]) -> None:
        # TODO contains duplicated code from the other add_match() methods
        if isinstance(match, RemoveBeforeMatch):
            self.front.errors[match.removed_sequence_length()][match.errors] += 1
        else:
            adjacent_base = match.adjacent_base()
            self.back.errors[match.removed_sequence_length()][match.errors] += 1
            try:
                self.back.adjacent_bases[adjacent_base] += 1
            except KeyError:
                self.back.adjacent_bases[""] += 1

    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
        return self.front, self.back


class Match(ABC):

    adapter: "Adapter"

    @abstractmethod
    def remainder_interval(self) -> Tuple[int, int]:
        pass

    @abstractmethod
    def retained_adapter_interval(self) -> Tuple[int, int]:
        pass

    @abstractmethod
    def get_info_records(self, read) -> List[List]:
        pass

    @abstractmethod
    def trimmed(self, read):
        pass

    @abstractmethod
    def match_sequence(self):
        pass


class SingleMatch(Match, ABC):
    """
    Representation of a single adapter matched to a single string
    """

    __slots__ = [
        "astart",
        "astop",
        "rstart",
        "rstop",
        "score",
        "errors",
        "adapter",
        "sequence",
        "length",
        "adjacent_base",
    ]

    def __init__(
        self,
        astart: int,
        astop: int,
        rstart: int,
        rstop: int,
        score: int,
        errors: int,
        adapter: "SingleAdapter",
        sequence: str,
    ):
        self.astart: int = astart
        self.astop: int = astop
        self.rstart: int = rstart
        self.rstop: int = rstop
        self.score: int = score
        self.errors: int = errors
        self.adapter: SingleAdapter = adapter
        self.sequence = sequence
        # Number of aligned characters in the adapter. If there are
        # indels, this may be different from the number of characters
        # in the read.
        self.length: int = astop - astart

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(astart={self.astart}, astop={self.astop}, "
            f"rstart={self.rstart}, rstop={self.rstop}, "
            f"score={self.score}, errors={self.errors})"
        )

    def __eq__(self, other) -> bool:
        return (
            other.__class__ is self.__class__
            and self.astart == other.astart
            and self.astop == other.astop
            and self.rstart == other.rstart
            and self.rstop == other.rstop
            and self.score == other.score
            and self.errors == other.errors
            and self.adapter is other.adapter
            and self.sequence == other.sequence
        )

    def wildcards(self, wildcard_char: str = "N") -> str:
        """
        Return a string that contains, for each wildcard character,
        the character that it matches. For example, if the adapter
        ATNGNA matches ATCGTA, then the string 'CT' is returned.

        If there are indels, this is not reliable as the full alignment
        is not available.
        """
        wildcards = [
            self.sequence[self.rstart + i]
            for i in range(self.length)
            if self.adapter.sequence[self.astart + i] == wildcard_char
            and self.rstart + i < len(self.sequence)
        ]
        return "".join(wildcards)

    def get_info_records(self, read) -> List[List]:
        seq = read.sequence
        qualities = read.qualities
        info = [
            "",
            self.errors,
            self.rstart,
            self.rstop,
            seq[0 : self.rstart],
            seq[self.rstart : self.rstop],
            seq[self.rstop :],
            self.adapter.name,
        ]
        if qualities:
            info += [
                qualities[0 : self.rstart],
                qualities[self.rstart : self.rstop],
                qualities[self.rstop :],
            ]
        else:
            info += ["", "", ""]

        return [info]

    def match_sequence(self):
        return self.sequence[self.rstart : self.rstop]

    @abstractmethod
    def removed_sequence_length(self) -> int:
        pass


class RemoveBeforeMatch(SingleMatch):
    """A match that removes sequence before the match"""

    def rest(self) -> str:
        """
        Return the part of the read before this match if this is a
        'front' (5') adapter,
        return the part after the match if this is not a 'front' adapter (3').
        This can be an empty string.
        """
        return self.sequence[: self.rstart]

    def remainder_interval(self) -> Tuple[int, int]:
        """
        Return an interval (start, stop) that describes the part of the read that would
        remain after trimming
        """
        return self.rstop, len(self.sequence)

    def retained_adapter_interval(self) -> Tuple[int, int]:
        return self.rstart, len(self.sequence)

    def trim_slice(self):
        # Same as remainder_interval, but as a slice() object
        return slice(self.rstop, None)

    def trimmed(self, read):
        return read[self.rstop :]

    def removed_sequence_length(self) -> int:
        return self.rstop


class RemoveAfterMatch(SingleMatch):
    """A match that removes sequence after the match"""

    def rest(self) -> str:
        """
        Return the part of the read before this match if this is a
        'front' (5') adapter,
        return the part after the match if this is not a 'front' adapter (3').
        This can be an empty string.
        """
        return self.sequence[self.rstop :]

    def remainder_interval(self) -> Tuple[int, int]:
        """
        Return an interval (start, stop) that describes the part of the read that would
        remain after trimming
        """
        return 0, self.rstart

    def retained_adapter_interval(self) -> Tuple[int, int]:
        return 0, self.rstop

    def trim_slice(self):
        # Same as remainder_interval, but as a slice() object
        return slice(None, self.rstart)

    def trimmed(self, read):
        return read[: self.rstart]

    def adjacent_base(self) -> str:
        return self.sequence[self.rstart - 1 : self.rstart]

    def removed_sequence_length(self) -> int:
        return len(self.sequence) - self.rstart


def _generate_adapter_name(_start=[1]) -> str:
    name = str(_start[0])
    _start[0] += 1
    return name


class Matchable(ABC):
    """Something that has a match_to() method."""

    def __init__(self, name: Optional[str], *args, **kwargs):
        self.name = name

    @abstractmethod
    def enable_debug(self):
        pass

    @abstractmethod
    def match_to(self, sequence: str):
        pass


class Adapter(Matchable, ABC):

    description = "adapter with one component"  # this is overriden in subclasses

    @abstractmethod
    def spec(self) -> str:
        """Return string representation of this adapter"""

    @abstractmethod
    def create_statistics(self) -> AdapterStatistics:
        pass

    @abstractmethod
    def descriptive_identifier(self) -> str:
        pass


class SingleAdapter(Adapter, ABC):
    """
    This class is used to find a single adapter characterized by sequence, error rate,
    type etc. within reads.

    Arguments:
        sequence (str): The adapter sequence. Will be converted to uppercase.
            Also, Us will be converted to Ts.

        max_errors: Maximum allowed errors (non-negative float). If the values is less than 1, this
            is interpreted as a rate and passed to the aligner. If it is 1 or greater, the value
            is converted to a rate by dividing it by the number of non-N characters in the sequence.

        The error rate is the number of errors in the alignment divided by the length
        of the part of the alignment that matches the adapter.

        min_overlap: Report a match only if at least this number of bases of the adapter match the
            read.

        read_wildcards: Whether IUPAC wildcards in the read are allowed.

        adapter_wildcards: Whether IUPAC wildcards in the adapter are allowed.

        name: Optional name of the adapter. If not provided, the name is set to a
            unique number.

        indels: Whether indels are allowed in the alignment.
    """

    allows_partial_matches: bool = True

    def __init__(
        self,
        sequence: str,
        max_errors: float = 0.1,
        min_overlap: int = 3,
        read_wildcards: bool = False,
        adapter_wildcards: bool = True,
        name: Optional[str] = None,
        indels: bool = True,
    ):
        self.name: str = _generate_adapter_name() if name is None else name
        super().__init__(self.name)
        self._debug: bool = False
        self.sequence: str = sequence.upper().replace("U", "T").replace("I", "N")
        if not self.sequence:
            raise ValueError("Adapter sequence is empty")
        if max_errors >= 1 and self.sequence.count("N") != len(self.sequence):
            max_errors /= len(self.sequence) - self.sequence.count("N")
        self.max_error_rate: float = max_errors
        self.min_overlap: int = min(min_overlap, len(self.sequence))
        iupac = frozenset("ABCDGHKMNRSTUVWXY")
        if adapter_wildcards and not set(self.sequence) <= iupac:
            for c in self.sequence:
                if c not in iupac:
                    raise InvalidCharacter(
                        f"Character '{c}' in adapter sequence '{self.sequence}' is "
                        f"not a valid IUPAC code. Use only characters 'ABCDGHIKMNRSTUVWXY'."
                    )
        # Optimization: Use non-wildcard matching if only ACGT is used
        self.adapter_wildcards: bool = adapter_wildcards and not set(
            self.sequence
        ) <= set("ACGT")
        self.read_wildcards: bool = read_wildcards
        self.indels: bool = indels
        self.aligner = self._aligner()

    def _make_aligner(self, flags: int) -> Aligner:
        # TODO
        # Indels are suppressed by setting their cost very high, but a different algorithm
        # should be used instead.
        indel_cost = 1 if self.indels else 100000
        return Aligner(
            self.sequence,
            self.max_error_rate,
            flags=flags,
            wildcard_ref=self.adapter_wildcards,
            wildcard_query=self.read_wildcards,
            indel_cost=indel_cost,
            min_overlap=self.min_overlap,
        )

    def __repr__(self):
        return (
            "<{cls}(name={name!r}, sequence={sequence!r}, "
            "max_error_rate={max_error_rate}, min_overlap={min_overlap}, "
            "read_wildcards={read_wildcards}, "
            "adapter_wildcards={adapter_wildcards}, "
            "indels={indels})>".format(cls=self.__class__.__name__, **vars(self))
        )

    @property
    def effective_length(self) -> int:
        return self.aligner.effective_length

    def enable_debug(self) -> None:
        """
        Print out the dynamic programming matrix after matching a read to an
        adapter.
        """
        self._debug = True
        self.aligner.enable_debug()

    @abstractmethod
    def _aligner(self):
        pass

    @abstractmethod
    def match_to(self, sequence: str):
        """
        Attempt to match this adapter to the given string.

        Return a Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """

    def __len__(self) -> int:
        return len(self.sequence)


class FrontAdapter(SingleAdapter):
    """A 5' adapter"""

    description = "regular 5'"

    def __init__(self, *args, **kwargs):
        self._force_anywhere = kwargs.pop("force_anywhere", False)
        super().__init__(*args, **kwargs)

    def descriptive_identifier(self) -> str:
        return "regular_five_prime"

    def _aligner(self) -> Aligner:
        return self._make_aligner(
            Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value
        )

    def match_to(self, sequence: str):
        """
        Attempt to match this adapter to the given read.

        Return a Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """
        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
            sequence
        )
        if self._debug:
            print(self.aligner.dpmatrix)
        if alignment is None:
            return None
        return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)

    def spec(self) -> str:
        return f"{self.sequence}..."

    def create_statistics(self) -> FrontAdapterStatistics:
        return FrontAdapterStatistics(self)


class RightmostFrontAdapter(FrontAdapter):
    """A 5' adapter that prefers rightmost matches"""

    description = "rightmost 5'"

    # def __init__(self, *args, **kwargs):
    #     self._force_anywhere = kwargs.pop("force_anywhere", False)
    #     super().__init__(*args, **kwargs)

    def descriptive_identifier(self) -> str:
        return "rightmost_five_prime"

    def _aligner(self) -> Aligner:
        self.sequence = self.sequence[::-1]
        aligner = self._make_aligner(
            Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value
        )
        self.sequence = self.sequence[::-1]
        return aligner

    def match_to(self, sequence: str):
        """
        Attempt to match this adapter to the given read.

        Return a Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """
        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
            sequence[::-1]
        )
        if self._debug:
            print(self.aligner.dpmatrix)
        if alignment is None:
            return None

        ref_start, ref_end, query_start, query_end, score, errors = alignment
        alignment = (
            len(self.sequence) - ref_end,
            len(self.sequence) - ref_start,
            len(sequence) - query_end,
            len(sequence) - query_start,
            score,
            errors,
        )
        return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)

    def spec(self) -> str:
        return f"{self.sequence}...;rightmost"


class BackAdapter(SingleAdapter):
    """A 3' adapter"""

    description = "regular 3'"

    def __init__(self, *args, **kwargs):
        self._force_anywhere = kwargs.pop("force_anywhere", False)
        super().__init__(*args, **kwargs)

    def descriptive_identifier(self) -> str:
        return "regular_three_prime"

    def _aligner(self):
        return self._make_aligner(
            Where.ANYWHERE.value if self._force_anywhere else Where.BACK.value
        )

    def match_to(self, sequence: str):
        """
        Attempt to match this adapter to the given read.

        Return a Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """
        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(
            sequence
        )
        if self._debug:
            print(self.aligner.dpmatrix)  # pragma: no cover
        if alignment is None:
            return None
        return RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)

    def spec(self) -> str:
        return f"{self.sequence}"

    def create_statistics(self) -> BackAdapterStatistics:
        return BackAdapterStatistics(self)


class AnywhereAdapter(SingleAdapter):
    """
    An adapter that can be 5' or 3'. If a match involves the first base of
    the read, it is assumed to be a 5' adapter and a 3' otherwise.
    """

    description = "variable 5'/3'"

    def descriptive_identifier(self) -> str:
        return "anywhere"

    def _aligner(self):
        return self._make_aligner(Where.ANYWHERE.value)

    def match_to(self, sequence: str):
        """
        Attempt to match this adapter to the given string.

        Return a Match instance if a match was found;
        return None if no match was found given the matching criteria (minimum
        overlap length, maximum error rate).
        """
        alignment = self.aligner.locate(sequence.upper())
        if self._debug:
            print(self.aligner.dpmatrix)
        if alignment is None:
            return None
        # guess: if alignment starts at pos 0, it’s a 5' adapter
        if alignment[2] == 0:  # index 2 is rstart
            match = RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore
        else:
            match = RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore
        return match

    def spec(self) -> str:
        return f"...{self.sequence}..."

    def create_statistics(self) -> AnywhereAdapterStatistics:
        return AnywhereAdapterStatistics(self)


class NonInternalFrontAdapter(FrontAdapter):
    """A non-internal 5' adapter"""

    description = "non-internal 5'"

    def descriptive_identifier(self) -> str:
        return "noninternal_five_prime"

    def _aligner(self):
        return self._make_aligner(Where.FRONT_NOT_INTERNAL.value)

    def match_to(self, sequence: str):
        # The locate function takes care of uppercasing the sequence
        alignment = self.aligner.locate(sequence)
        if self._debug:
            try:
                print(self.aligner.dpmatrix)
            except AttributeError:
                pass
        if alignment is None:
            return None
        return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore

    def spec(self) -> str:
        return f"X{self.sequence}..."


class NonInternalBackAdapter(BackAdapter):
    """A non-internal 3' adapter"""

    description = "non-internal 3'"

    def descriptive_identifier(self) -> str:
        return "noninternal_three_prime"

    def _aligner(self):
        return self._make_aligner(Where.BACK_NOT_INTERNAL.value)

    def match_to(self, sequence: str):
        # The locate function takes care of uppercasing the sequence
        alignment = self.aligner.locate(sequence)
        if self._debug:
            try:
                print(self.aligner.dpmatrix)  # pragma: no cover
            except AttributeError:
                pass
        if alignment is None:
            return None
        return RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore

    def spec(self) -> str:
        return f"{self.sequence}X"


class PrefixAdapter(NonInternalFrontAdapter):
    """An anchored 5' adapter"""

    description = "anchored 5'"
    allows_partial_matches = False

    def __init__(self, sequence: str, *args, **kwargs):
        kwargs["min_overlap"] = len(sequence)
        super().__init__(sequence, *args, **kwargs)

    def descriptive_identifier(self) -> str:
        return "anchored_five_prime"

    def _aligner(self):
        if not self.indels:  # TODO or if error rate allows 0 errors anyway
            return PrefixComparer(
                self.sequence,
                self.max_error_rate,
                wildcard_ref=self.adapter_wildcards,
                wildcard_query=self.read_wildcards,
                min_overlap=self.min_overlap,
            )
        else:
            return self._make_aligner(Where.PREFIX.value)

    def spec(self) -> str:
        return f"^{self.sequence}..."


class SuffixAdapter(NonInternalBackAdapter):
    """An anchored 3' adapter"""

    description = "anchored 3'"
    allows_partial_matches = False

    def __init__(self, sequence: str, *args, **kwargs):
        kwargs["min_overlap"] = len(sequence)
        super().__init__(sequence, *args, **kwargs)

    def descriptive_identifier(self) -> str:
        return "anchored_three_prime"

    def _aligner(self):
        if not self.indels:  # TODO or if error rate allows 0 errors anyway
            return SuffixComparer(
                self.sequence,
                self.max_error_rate,
                wildcard_ref=self.adapter_wildcards,
                wildcard_query=self.read_wildcards,
                min_overlap=self.min_overlap,
            )
        else:
            return self._make_aligner(Where.SUFFIX.value)

    def spec(self) -> str:
        return f"{self.sequence}$"


class LinkedMatch(Match):
    """
    Represent a match of a LinkedAdapter
    """

    def __init__(
        self,
        front_match: RemoveBeforeMatch,
        back_match: RemoveAfterMatch,
        adapter: "LinkedAdapter",
    ):
        assert front_match is not None or back_match is not None
        self.front_match: RemoveBeforeMatch = front_match
        self.back_match: RemoveAfterMatch = back_match
        self.adapter: LinkedAdapter = adapter

    def __repr__(self):
        return "<LinkedMatch(front_match={!r}, back_match={}, adapter={})>".format(
            self.front_match, self.back_match, self.adapter
        )

    @property
    def score(self):
        """Number of matching bases"""
        s = 0
        if self.front_match is not None:
            s += self.front_match.score
        if self.back_match is not None:
            s += self.back_match.score
        return s

    @property
    def errors(self):
        e = 0
        if self.front_match is not None:
            e += self.front_match.errors
        if self.back_match is not None:
            e += self.back_match.errors
        return e

    def trimmed(self, read):
        if self.front_match:
            read = self.front_match.trimmed(read)
        if self.back_match:
            read = self.back_match.trimmed(read)
        return read

    def remainder_interval(self) -> Tuple[int, int]:
        matches = [
            match for match in [self.front_match, self.back_match] if match is not None
        ]
        return remainder(matches)

    def retained_adapter_interval(self) -> Tuple[int, int]:
        if self.front_match:
            start = self.front_match.rstart
            offset = self.front_match.rstop
        else:
            start = offset = 0
        if self.back_match:
            end = self.back_match.rstop + offset
        else:
            end = len(self.front_match.sequence)
        return start, end

    def get_info_records(self, read) -> List[List]:
        records = []
        for match, namesuffix in [
            (self.front_match, ";1"),
            (self.back_match, ";2"),
        ]:
            if match is None:
                continue
            record = match.get_info_records(read)[0]
            record[7] = (
                "none" if self.adapter.name is None else self.adapter.name
            ) + namesuffix
            records.append(record)
            read = match.trimmed(read)
        return records

    def match_sequence(self):
        return (
            (self.front_match.match_sequence() if self.front_match else "")
            + ","
            + (self.back_match.match_sequence() if self.back_match else "")
        )


class LinkedAdapter(Adapter):
    """A 5' adapter combined with a 3' adapter"""

    description = "linked"

    def __init__(
        self,
        front_adapter: SingleAdapter,
        back_adapter: SingleAdapter,
        front_required: bool,
        back_required: bool,
        name: Optional[str],
    ):
        super().__init__(name)
        self.front_required = front_required
        self.back_required = back_required

        # The following attributes are needed for the report
        self.where = "linked"
        self.name: str = _generate_adapter_name() if name is None else name
        self.front_adapter = front_adapter
        self.front_adapter.name = self.name
        self.back_adapter = back_adapter

    def descriptive_identifier(self) -> str:
        return "linked"

    def enable_debug(self):
        self.front_adapter.enable_debug()
        self.back_adapter.enable_debug()

    def match_to(self, sequence: str) -> Optional[LinkedMatch]:
        """
        Match the two linked adapters against a string
        """
        front_match = self.front_adapter.match_to(sequence)
        if self.front_required and front_match is None:
            return None
        if front_match is not None:
            sequence = sequence[front_match.trim_slice()]
        back_match = self.back_adapter.match_to(sequence)
        if back_match is None and (self.back_required or front_match is None):
            return None
        return LinkedMatch(front_match, back_match, self)

    def create_statistics(self) -> LinkedAdapterStatistics:
        return LinkedAdapterStatistics(
            self, front=self.front_adapter, back=self.back_adapter
        )

    @property
    def sequence(self):
        return self.front_adapter.sequence + "..." + self.back_adapter.sequence

    @property
    def remove(self):
        return None

    def spec(self) -> str:
        return f"{self.front_adapter.spec()}...{self.back_adapter.spec()}"


class MultipleAdapters(Matchable):
    """
    Represent multiple adapters at once
    """

    def __init__(self, adapters: Sequence[Matchable]):
        super().__init__(name="multiple_adapters")
        self._adapters = adapters

    def enable_debug(self):
        for a in self._adapters:
            a.enable_debug()

    def __getitem__(self, item):
        return self._adapters[item]

    def __len__(self):
        return len(self._adapters)

    def match_to(self, sequence: str) -> Optional[SingleMatch]:
        """
        Find the adapter that best matches the sequence.

        Return either a Match instance or None if there are no matches.
        """
        best_match = None
        for adapter in self._adapters:
            match = adapter.match_to(sequence)
            if match is None:
                continue

            # the score determines which adapter fits best
            if (
                best_match is None
                or match.score > best_match.score
                or (
                    match.score == best_match.score and match.errors < best_match.errors
                )
            ):
                best_match = match
        return best_match


class IndexedAdapters(Matchable, ABC):
    """
    Represent multiple adapters of the same type at once and use an index data structure
    to speed up matching. This acts like a "normal" Adapter as it provides a match_to
    method, but is faster with lots of adapters.

    There are quite a few restrictions:
    - the error rate allows at most 2 mismatches
    - wildcards in the adapter are not allowed
    - wildcards in the read are not allowed

    Use the is_acceptable() method to check individual adapters.
    """

    AdapterIndex = Dict[str, Tuple[SingleAdapter, int, int]]

    def __init__(self, adapters):
        """All given adapters must be of the same type"""
        super().__init__(name="indexed_adapters")
        if not adapters:
            raise ValueError("Adapter list is empty")
        for adapter in adapters:
            self._accept(adapter)
        self._adapters = adapters
        self._multiple_adapters = MultipleAdapters(adapters)
        self._lengths, self._index = self._make_index()
        logger.debug(
            "String lengths in the index: %s", sorted(self._lengths, reverse=True)
        )
        if len(self._lengths) == 1:
            self._length = self._lengths[0]
            self.match_to = self._match_to_one_length
        else:
            self.match_to = self._match_to_multiple_lengths
        self._make_affix = self._get_make_affix()

    def __repr__(self):
        return f"{self.__class__.__name__}(adapters={self._adapters!r})"

    def match_to(self, sequence: str):
        """Never called because it gets overwritten in __init__"""

    @abstractmethod
    def _get_make_affix(self):
        pass

    @abstractmethod
    def _make_match(self, adapter, length, matches, errors, sequence) -> SingleMatch:
        pass

    @classmethod
    def _accept(cls, adapter):
        """Raise a ValueError if the adapter is not acceptable"""
        if adapter.read_wildcards:
            raise ValueError("Wildcards in the read not supported")
        if adapter.adapter_wildcards:
            raise ValueError("Wildcards in the adapter not supported")
        k = int(len(adapter) * adapter.max_error_rate)
        if k > 2:
            raise ValueError("Error rate too high")

    @classmethod
    def is_acceptable(cls, adapter):
        """
        Return whether this adapter is acceptable for being used in an index

        Adapters are not acceptable if they allow wildcards, allow too many errors,
        or would lead to a very large index.
        """
        try:
            cls._accept(adapter)
        except ValueError:
            return False
        return True

    def _make_index(self) -> Tuple[List[int], "AdapterIndex"]:
        logger.info("Building index of %s adapters ...", len(self._adapters))
        index: Dict[str, Tuple[SingleAdapter, int, int]] = dict()
        lengths = set()
        has_warned = False
        for adapter in self._adapters:
            sequence = adapter.sequence
            k = int(adapter.max_error_rate * len(sequence))
            environment = edit_environment if adapter.indels else hamming_environment
            for s, errors, matches in environment(sequence, k):
                if s in index:
                    other_adapter, other_errors, other_matches = index[s]
                    if matches < other_matches:
                        continue
                    if other_matches == matches and not has_warned:
                        logger.warning(
                            "Adapters %s %r and %s %r are very similar. At %s allowed errors, "
                            "the sequence %r cannot be assigned uniquely because the number of "
                            "matches is %s compared to both adapters.",
                            other_adapter.name,
                            other_adapter.sequence,
                            adapter.name,
                            adapter.sequence,
                            k,
                            s,
                            matches,
                        )
                        has_warned = True
                else:
                    index[s] = (adapter, errors, matches)
                lengths.add(len(s))
        logger.info("Built an index containing %s strings.", len(index))

        return sorted(lengths, reverse=True), index

    def _match_to_one_length(self, sequence: str):
        """
        Match the adapters against a string and return a Match that represents
        the best match or None if no match was found
        """
        affix = self._make_affix(sequence.upper(), self._length)
        if "N" in affix:
            # Fall back to non-indexed matching
            return self._multiple_adapters.match_to(sequence)
        try:
            adapter, e, m = self._index[affix]
        except KeyError:
            return None
        return self._make_match(adapter, self._length, m, e, sequence)

    def _match_to_multiple_lengths(self, sequence: str):
        """
        Match the adapters against a string and return a Match that represents
        the best match or None if no match was found
        """
        affix = sequence.upper()

        # Check all the prefixes or suffixes (affixes) that could match
        best_adapter: Optional[SingleAdapter] = None
        best_length = 0
        best_m = -1
        best_e = 1000
        check_n = True
        for length in self._lengths:
            if length < best_m:
                # No chance of getting the same or a higher number of matches, so we can stop early
                break
            affix = self._make_affix(affix, length)
            if check_n:
                if "N" in affix:
                    return self._multiple_adapters.match_to(sequence)
                check_n = False
            try:
                adapter, e, m = self._index[affix]
            except KeyError:
                continue
            if m > best_m or (m == best_m and e < best_e):
                # TODO this could be made to work:
                # assert best_m == -1
                best_adapter = adapter
                best_e = e
                best_m = m
                best_length = length

        if best_m == -1:
            return None
        else:
            return self._make_match(best_adapter, best_length, best_m, best_e, sequence)

    def enable_debug(self):
        pass


class IndexedPrefixAdapters(IndexedAdapters):
    @classmethod
    def _accept(cls, adapter):
        if not isinstance(adapter, PrefixAdapter):
            raise ValueError("Only 5' anchored adapters are allowed")
        return super()._accept(adapter)

    def _make_match(self, adapter, length, score, errors, sequence):
        return RemoveBeforeMatch(
            astart=0,
            astop=len(adapter.sequence),
            rstart=0,
            rstop=length,
            score=score,
            errors=errors,
            adapter=adapter,
            sequence=sequence,
        )

    def _get_make_affix(self):
        return self._make_prefix

    @staticmethod
    def _make_prefix(s, n):
        return s[:n]


class IndexedSuffixAdapters(IndexedAdapters):
    @classmethod
    def _accept(cls, adapter):
        if not isinstance(adapter, SuffixAdapter):
            raise ValueError("Only anchored 3' adapters are allowed")
        return super()._accept(adapter)

    def _make_match(self, adapter, length, score, errors, sequence):
        return RemoveAfterMatch(
            astart=0,
            astop=len(adapter.sequence),
            rstart=len(sequence) - length,
            rstop=len(sequence),
            score=score,
            errors=errors,
            adapter=adapter,
            sequence=sequence,
        )

    def _get_make_affix(self):
        return self._make_suffix

    @staticmethod
    def _make_suffix(s, n):
        return s[-n:]


def warn_duplicate_adapters(adapters):
    d = dict()
    for adapter in adapters:
        key = (adapter.__class__, adapter.sequence)
        if key in d:
            logger.warning(
                "Adapter %r (%s) was specified multiple times! "
                "Please make sure that this is what you want.",
                adapter.sequence,
                adapter.description,
            )
        d[key] = adapter.name


def remainder(matches: Sequence[Match]) -> Tuple[int, int]:
    """
    Determine which section of the read would not be trimmed. Return a tuple (start, stop)
    that gives the interval of the untrimmed part relative to the original read.

    matches must be non-empty
    """
    if not matches:
        raise ValueError("matches must not be empty")
    start = 0
    for match in matches:
        match_start, match_stop = match.remainder_interval()
        start += match_start
    length = match_stop - match_start
    return (start, start + length)



from cutadapt.statistics import ReadLengthStatistics


class TestReadLengthStatistics:
    def test_empty_on_init(self):
        rls = ReadLengthStatistics()
        assert rls.written_reads() == 0
        assert rls.written_bp() == (0, 0)
        lengths = rls.written_lengths()
        assert not lengths[0] and not lengths[1]

    def test_some_reads(self):
        rls = ReadLengthStatistics()
        rls.update("THEREAD")  # length: 7
        rls.update("YETANOTHER")  # length: 10
        rls.update2("FIRST", "SECOND")  # lengths: 5, 6
        rls.update("12345")

        assert rls.written_reads() == 4
        assert rls.written_bp() == (7 + 10 + 5 + 5, 6)
        lengths = rls.written_lengths()
        assert sorted(lengths[0].items()) == [(5, 2), (7, 1), (10, 1)]
        assert sorted(lengths[1].items()) == [(6, 1)]

    def test_iadd(self):
        rls = ReadLengthStatistics()
        rls.update("THEREAD")  # length: 7
        rls.update("YETANOTHER")  # length: 10
        rls.update2("FIRST", "SECOND")  # lengths: 5, 6
        rls.update("12345")

        rls2 = ReadLengthStatistics()
        rls2.update("TESTING")  # length: 7
        rls2.update2("LEFT", "RIGHT")  # lengths: 4, 5
        rls += rls2

        assert rls.written_reads() == 6
        assert rls.written_bp() == (7 + 10 + 5 + 5 + 7 + 4, 6 + 5)
        lengths = rls.written_lengths()
        assert sorted(lengths[0].items()) == [(4, 1), (5, 2), (7, 2), (10, 1)]
        assert sorted(lengths[1].items()) == [(5, 1), (6, 1)]



"""
Cutadapt doesn’t have a stable API, yet. This is an attempt to document how
one currently needs to use Cutadapt from Python to do certain things,
mostly in order to figure out where improvements need to be made.

The tests in this module do not check results, they are just here to
ensure that the code as shown can be executed.
"""
import copy
import json
import os

from utils import datapath


def test_command_line():
    # Call Cutadapt from Python, but pass parameters as a list of strings
    # the same way we would in the shell. The difference is that this is
    # not in a separate process, errors cause a CommandLineError instead
    # of SystemExit, and we get back a Statistics object.
    from cutadapt.__main__ import main

    stats = main(["-q", "10", "-o", os.devnull, datapath("small.fastq")])
    assert stats is not None
    json.dumps(stats.as_json())

    # TODO
    # - Should not set up logging
    # - Should not print anything
    # - still raises SystemExit if parser.error is called
    # - Should be cutadapt.run(...)
    # - Should the JSON stats be returned instead?


def test_pipeline_single(tmp_path):
    # The following is roughly equivalent to:
    # cutadapt -u 5 -a GATCGGAAGA -q 0,15 -m 10
    #   --discard-untrimmed --info-file=info.txt -o ... small.fastq
    import json
    from cutadapt.pipeline import SingleEndPipeline
    from cutadapt.utils import FileOpener
    from cutadapt.modifiers import UnconditionalCutter, QualityTrimmer, AdapterCutter
    from cutadapt.adapters import BackAdapter
    from cutadapt.pipeline import InputPaths, OutputFiles

    file_opener = FileOpener()
    adapter = BackAdapter(
        sequence="GATCGGAAGA",
        max_errors=1,
        min_overlap=3,
    )
    modifiers = [
        UnconditionalCutter(5),
        QualityTrimmer(cutoff_front=0, cutoff_back=15),
        AdapterCutter([adapter]),
    ]
    pipeline = SingleEndPipeline(modifiers)
    pipeline.minimum_length = (10,)
    pipeline.discard_untrimmed = True
    inpaths = InputPaths(datapath("small.fastq"))
    info_file = file_opener.xopen_or_none(tmp_path / "info.txt", "wb")
    out = file_opener.xopen(tmp_path / "out.fastq", "wb")
    outfiles = OutputFiles(info=info_file, out=out)
    stats = pipeline.run(inpaths, outfiles, cores=1)
    assert stats is not None
    json.dumps(stats.as_json())
    outfiles.close()
    # TODO
    # - info file isn’t written, what is missing?
    # - see next function for more TODOs that also apply here


def test_pipeline_paired(tmp_path):
    # cutadapt -u 5 -U 7 -a GATCGGAAGA -q 0,15 -m 10:0
    #   --discard-untrimmed --info-file=info.txt
    #   -o ... -p ...
    #   paired.1.fastq paired.2.fastq
    from cutadapt.pipeline import PairedEndPipeline
    from cutadapt.utils import FileOpener
    from cutadapt.modifiers import UnconditionalCutter, QualityTrimmer, AdapterCutter
    from cutadapt.adapters import BackAdapter
    from cutadapt.pipeline import InputPaths, OutputFiles

    trimmer = QualityTrimmer(cutoff_front=0, cutoff_back=15)
    adapter = BackAdapter(
        sequence="GATCGGAAGA",
        max_errors=1,
        min_overlap=3,
    )
    modifiers = [
        (UnconditionalCutter(5), UnconditionalCutter(7)),
        (trimmer, copy.copy(trimmer)),
        (AdapterCutter([adapter]), None),
    ]

    pipeline = PairedEndPipeline(modifiers, "any")

    pipeline.minimum_length = (10, None)
    pipeline.discard_untrimmed = True

    file_opener = FileOpener()
    inpaths = InputPaths(datapath("paired.1.fastq"), datapath("paired.2.fastq"))
    info_file = file_opener.xopen_or_none(tmp_path / "info.txt", "wb")
    out, out2 = file_opener.xopen_pair(
        tmp_path / "out.1.fastq", tmp_path / "out.2.fastq", "wb"
    )
    outfiles = OutputFiles(
        info=info_file,
        out=out,
        out2=out2,
    )
    stats = pipeline.run(inpaths, outfiles, cores=1, progress=True)
    assert stats is not None
    _ = stats.as_json()
    outfiles.close()

    # TODO
    # - could use += for adding modifiers
    # - allow using adapter specification strings?
    # - filters as attributes on Pipeline is awkward
    # - too many submodules (flatter namespace)
    # - more default arguments
    # - info file isn’t written, what is missing?
    # - use xopen directly instead of file_opener;
    #   possibly with myxopen = functools.partial(xopen, ...)


# How this could look in the future:
# def test_pipeline_single_new(tmp_path):
#     import json
#     from cutadapt.pipeline import SingleEndPipeline
#     from cutadapt.modifiers import UnconditionalCutter, QualityTrimmer, AdapterCutter
#     from cutadapt.adapters import BackAdapter
#     from cutadapt.steps import InfoFileWriter, SingleEndFilter, SingleEndSink
#     from cutadapt.filters import TooShort, DiscardUntrimmed
#     from contextlib import ExitStack
#     from xopen import xopen
#
#     with ExitStack() as stack:
#         # Input files
#         infile = stack.enter_context(xopen(datapath("small.fastq"), "rb"))
#
#         # Output files
#         info_file = stack.enter_context(xopen(tmp_path / "info.txt", "wb"))
#         out = stack.enter_context(xopen(tmp_path / "out.fastq", "wb"))
#         too_short = stack.enter_context(xopen(tmp_path / "tooshort.fastq", "wb"))
#         pipeline = SingleEndPipeline(
#             modifiers=[
#                 UnconditionalCutter(5),
#                 QualityTrimmer(cutoff_front=0, cutoff_back=15),
#                 AdapterCutter([BackAdapter(sequence="GATCGGAAGA", max_errors=1, min_overlap=3)]),
#             ],
#             filters=[
#                 InfoFileWriter(info_file),
#                 SingleEndFilter(TooShort(minimum_length=10), too_short),
#                 SingleEndFilter(DiscardUntrimmed()),
#             ],
#             sink=SingleEndSink(out),
#         )
#         stats = pipeline.run(infile, workers=1, progress=True)
#     json.dumps(stats.as_json())



"""
Filtering criteria (predicates)
"""
from abc import ABC, abstractmethod

from .qualtrim import expected_errors
from .modifiers import ModificationInfo


class Predicate(ABC):
    @abstractmethod
    def test(self, read, info: ModificationInfo) -> bool:
        """
        Return True if the filtering criterion matches.
        """

    @classmethod
    def descriptive_identifier(cls) -> str:
        """
        Return a short name for this predicate based on the class name such as "too_long",
        "too_many_expected_errors".
        This is used as identifier in the JSON report.
        """
        return "".join(
            ("_" + ch.lower() if ch.isupper() else ch) for ch in cls.__name__
        )[1:]


class TooShort(Predicate):
    """Select reads that are shorter than the specified minimum length"""

    def __init__(self, minimum_length: int):
        self.minimum_length = minimum_length

    def __repr__(self):
        return f"TooShort(minimum_length={self.minimum_length})"

    def test(self, read, info: ModificationInfo):
        return len(read) < self.minimum_length


class TooLong(Predicate):
    """Select reads that are longer than the specified maximum length"""

    def __init__(self, maximum_length: int):
        self.maximum_length = maximum_length

    def __repr__(self):
        return f"TooLong(maximum_length={self.maximum_length})"

    def test(self, read, info: ModificationInfo):
        return len(read) > self.maximum_length


class TooManyExpectedErrors(Predicate):
    """
    Select reads whose expected number of errors, according to the quality
    values, exceeds a threshold.

    The idea comes from usearch's -fastq_maxee parameter
    (http://drive5.com/usearch/).
    """

    def __init__(self, max_errors: float):
        self.max_errors = max_errors

    def __repr__(self):
        return f"TooManyExpectedErrors(max_errors={self.max_errors})"

    def test(self, read, info: ModificationInfo):
        return expected_errors(read.qualities) > self.max_errors


class TooManyN(Predicate):
    """
    Select reads that have too many 'N' bases.

    Both a raw count or a proportion (relative to the sequence length) can be used.
    """

    def __init__(self, count: float):
        """
        count -- if it is below 1.0, it will be considered a proportion, and above and equal to
        1 will be considered as discarding reads with a number of N's greater than this cutoff.
        """
        assert count >= 0
        self.is_proportion = count < 1.0
        self.cutoff = count

    def __repr__(self):
        return f"TooManyN(cutoff={self.cutoff}, is_proportion={self.is_proportion})"

    def test(self, read, info: ModificationInfo):
        n_count = read.sequence.lower().count("n")
        if self.is_proportion:
            if len(read) == 0:
                return False
            return n_count / len(read) > self.cutoff
        else:
            return n_count > self.cutoff


class CasavaFiltered(Predicate):
    """
    Select reads that have failed the CASAVA filter according to the read header.
    The headers look like ``xxxx x:Y:x:x`` (with a ``Y``). Reads that pass the filter
    have an ``N`` instead of ``Y``.

    Reads with unrecognized headers are not selected.
    """

    def __repr__(self):
        return "CasavaFiltered()"

    def test(self, read, info: ModificationInfo):
        _, _, right = read.name.partition(" ")
        return right[1:4] == ":Y:"  # discard if :Y: found


class DiscardUntrimmed(Predicate):
    """
    Select reads for which no adapter match was found
    """

    def __repr__(self):
        return "DiscardUntrimmed()"

    def test(self, read, info: ModificationInfo):
        return not info.matches


class DiscardTrimmed(Predicate):
    """
    Select reads for which at least one adapter match was found
    """

    def __repr__(self):
        return "DiscardTrimmed()"

    def test(self, read, info: ModificationInfo):
        return bool(info.matches)



"""
Routines for printing a report.
"""
import sys
from dataclasses import dataclass
from io import StringIO
import textwrap
from collections import defaultdict
from typing import Any, Optional, List, Dict, Iterator, Tuple
from .adapters import (
    EndStatistics,
    AdapterStatistics,
    FrontAdapter,
    BackAdapter,
    AnywhereAdapter,
    LinkedAdapter,
    SingleAdapter,
    LinkedAdapterStatistics,
    FrontAdapterStatistics,
    BackAdapterStatistics,
    AnywhereAdapterStatistics,
)
from .json import OneLine
from .modifiers import (
    QualityTrimmer,
    NextseqQualityTrimmer,
    AdapterCutter,
    PairedAdapterCutter,
    ReverseComplementer,
    PairedEndModifierWrapper,
)
from .statistics import ReadLengthStatistics
from .steps import SingleEndFilter, PairedEndFilter, HasStatistics

FILTERS = {
    "too_short": "that were too short",
    "too_long": "that were too long",
    "too_many_n": "with too many N",
    "too_many_expected_errors": "with too many exp. errors",
    "casava_filtered": "failed CASAVA filter",
    "discard_trimmed": "discarded as trimmed",
    "discard_untrimmed": "discarded as untrimmed",
}


def safe_divide(numerator: Optional[int], denominator: int) -> float:
    if numerator is None or not denominator:
        return 0.0
    else:
        return numerator / denominator


def add_if_not_none(a: Optional[int], b: Optional[int]) -> Optional[int]:
    if a is None:
        return b
    if b is None:
        return a
    return a + b


class Statistics:
    def __init__(self) -> None:
        """ """
        self.paired: Optional[bool] = None
        # Map a filter name to the number of filtered reads/read pairs
        self.filtered: Dict[str, int] = defaultdict(int)
        self.reverse_complemented: Optional[int] = None
        self.n = 0
        self.total_bp = [0, 0]
        self.read_length_statistics = ReadLengthStatistics()
        self.with_adapters: List[Optional[int]] = [None, None]
        self.quality_trimmed_bp: List[Optional[int]] = [None, None]
        self.adapter_stats: List[List[AdapterStatistics]] = [[], []]
        self._collected: bool = False

    def __iadd__(self, other: Any):
        if not isinstance(other, Statistics):
            raise ValueError(f"Cannot add {other.__type__.__name__}")
        self.n += other.n
        self.read_length_statistics += other.read_length_statistics

        if self.paired is None:
            self.paired = other.paired
        elif self.paired != other.paired:
            raise ValueError("Incompatible Statistics: paired is not equal")

        self.reverse_complemented = add_if_not_none(
            self.reverse_complemented, other.reverse_complemented
        )

        for filter_name, count in other.filtered.items():
            self.filtered[filter_name] += count

        for i in (0, 1):
            self.total_bp[i] += other.total_bp[i]
            self.with_adapters[i] = add_if_not_none(
                self.with_adapters[i], other.with_adapters[i]
            )
            self.quality_trimmed_bp[i] = add_if_not_none(
                self.quality_trimmed_bp[i], other.quality_trimmed_bp[i]
            )
            if self.adapter_stats[i] and other.adapter_stats[i]:
                if len(self.adapter_stats[i]) != len(other.adapter_stats[i]):
                    raise ValueError(
                        "Incompatible Statistics objects (adapter_stats length)"
                    )
                for j in range(len(self.adapter_stats[i])):
                    self.adapter_stats[i][j] += other.adapter_stats[i][j]
            elif other.adapter_stats[i]:
                assert self.adapter_stats[i] == []
                self.adapter_stats[i] = other.adapter_stats[i]
        return self

    def collect(
        self, n: int, total_bp1: int, total_bp2: Optional[int], modifiers, steps
    ):
        """
        n -- total number of reads
        total_bp1 -- number of bases in first reads
        total_bp2 -- number of bases in second reads. None for single-end data.
        """
        if self._collected:
            raise ValueError("Cannot call Statistics.collect more than once")
        self.n = n
        self.total_bp[0] = total_bp1
        if total_bp2 is None:
            self.paired = False
        else:
            self.paired = True
            self.total_bp[1] = total_bp2

        for step in steps:
            self._collect_step(step)
        for modifier in modifiers:
            self._collect_modifier(modifier)
        self._collected = True

        # For chaining
        return self

    def _collect_step(self, step) -> None:
        if isinstance(step, HasStatistics):
            self.read_length_statistics += step.get_statistics()
        if isinstance(step, (SingleEndFilter, PairedEndFilter)):
            name = step.descriptive_identifier()
            self.filtered[name] = step.filtered

    def _collect_modifier(self, m) -> None:
        if isinstance(m, PairedAdapterCutter):
            for i in 0, 1:
                self.with_adapters[i] = m.with_adapters
                self.adapter_stats[i] = list(m.adapter_statistics[i].values())
            return
        if isinstance(m, PairedEndModifierWrapper):
            modifiers_list = [(0, m._modifier1), (1, m._modifier2)]
        else:
            modifiers_list = [(0, m)]
        for i, modifier in modifiers_list:
            if isinstance(modifier, (QualityTrimmer, NextseqQualityTrimmer)):
                self.quality_trimmed_bp[i] = add_if_not_none(
                    self.quality_trimmed_bp[i], modifier.trimmed_bases
                )
            elif isinstance(modifier, AdapterCutter):
                assert self.with_adapters[i] is None
                self.with_adapters[i] = modifier.with_adapters
                self.adapter_stats[i] = list(modifier.adapter_statistics.values())
            elif isinstance(modifier, ReverseComplementer):
                assert self.with_adapters[i] is None
                self.with_adapters[i] = modifier.adapter_cutter.with_adapters
                self.adapter_stats[i] = list(
                    modifier.adapter_cutter.adapter_statistics.values()
                )
                self.reverse_complemented = modifier.reverse_complemented

    def as_json(self, gc_content: float = 0.5, one_line: bool = False) -> Dict:
        """
        Return a dict representation suitable for dumping in JSON format

        To achieve a more compact representation, set one_line to True, which
        will wrap some items in a `cutadapt.json.OneLine` object, and use
        `cutadapt.json.dumps` instead of `json.dumps` to dump the dict.
        """
        filtered = {name: self.filtered.get(name) for name in FILTERS.keys()}
        filtered_total = sum(self.filtered.values())
        written_reads = self.read_length_statistics.written_reads()
        written_bp = self.read_length_statistics.written_bp()
        assert written_reads + filtered_total == self.n
        return {
            "read_counts": {  # pairs or reads
                "input": self.n,
                "filtered": filtered,
                "output": self.read_length_statistics.written_reads(),
                "reverse_complemented": self.reverse_complemented,
                "read1_with_adapter": self.with_adapters[0],
                "read2_with_adapter": self.with_adapters[1] if self.paired else None,
            },
            "basepair_counts": {
                "input": self.total,
                "input_read1": self.total_bp[0],
                "input_read2": self.total_bp[1] if self.paired else None,
                "quality_trimmed": self.quality_trimmed,
                "quality_trimmed_read1": self.quality_trimmed_bp[0],
                "quality_trimmed_read2": self.quality_trimmed_bp[1],
                "output": self.total_written_bp,
                "output_read1": written_bp[0],
                "output_read2": written_bp[1] if self.paired else None,
            },
            "adapters_read1": [
                self._adapter_statistics_as_json(
                    astats, self.n, gc_content, one_line=one_line
                )
                for astats in self.adapter_stats[0]
            ],
            "adapters_read2": [
                self._adapter_statistics_as_json(
                    astats, self.n, gc_content, one_line=one_line
                )
                for astats in self.adapter_stats[1]
            ]
            if self.paired
            else None,
        }

    def _adapter_statistics_as_json(
        self,
        adapter_statistics: AdapterStatistics,
        n: int,
        gc_content: float,
        one_line: bool = False,
    ):
        adapter = adapter_statistics.adapter
        ends: List[Optional[Dict[str, Any]]] = []
        total_trimmed_reads = 0
        make_line = OneLine if one_line else list
        for end_statistics in adapter_statistics.end_statistics():
            if end_statistics is None:
                ends.append(None)
                continue
            total = sum(end_statistics.lengths.values())
            if end_statistics.allows_partial_matches:
                eranges = ErrorRanges(
                    length=end_statistics.effective_length,
                    error_rate=end_statistics.max_error_rate,
                ).lengths()
            else:
                eranges = None
            base_stats = AdjacentBaseStatistics(end_statistics.adjacent_bases)
            trimmed_lengths = [
                make_line(
                    {
                        "len": row.length,
                        "expect": round(row.expect, 1),
                        "counts": row.error_counts,
                    }
                )
                for row in histogram_rows(end_statistics, n, gc_content)
            ]
            ends.append(
                {
                    "type": end_statistics.adapter_type,
                    "sequence": end_statistics.sequence,
                    "error_rate": end_statistics.max_error_rate,
                    "indels": end_statistics.indels,
                    "error_lengths": make_line(eranges),
                    "matches": total,
                    "adjacent_bases": base_stats.as_json(),
                    "dominant_adjacent_base": base_stats.warnbase,
                    "trimmed_lengths": trimmed_lengths,
                }
            )
            total_trimmed_reads += total

        on_reverse_complement = (
            adapter_statistics.reverse_complemented
            if self.reverse_complemented
            else None
        )
        return {
            "name": adapter_statistics.name,
            "total_matches": total_trimmed_reads,
            "on_reverse_complement": on_reverse_complement,
            "linked": isinstance(adapter, LinkedAdapter),
            "five_prime_end": ends[0],
            "three_prime_end": ends[1],
        }

    @property
    def total(self) -> int:
        return sum(self.total_bp)

    @property
    def quality_trimmed(self) -> Optional[int]:
        return add_if_not_none(*self.quality_trimmed_bp)

    @property
    def total_written_bp(self) -> int:
        return sum(self.read_length_statistics.written_bp())

    @property
    def written(self) -> int:
        return self.read_length_statistics.written_reads()

    @property
    def written_fraction(self) -> float:
        return safe_divide(self.read_length_statistics.written_reads(), self.n)

    @property
    def with_adapters_fraction(self) -> List[float]:
        return [safe_divide(v, self.n) for v in self.with_adapters]

    @property
    def quality_trimmed_fraction(self) -> float:
        return safe_divide(self.quality_trimmed, self.total)

    @property
    def written_bp(self) -> Tuple[int, int]:
        return self.read_length_statistics.written_bp()

    @property
    def total_written_bp_fraction(self) -> float:
        return safe_divide(self.total_written_bp, self.total)

    @property
    def reverse_complemented_fraction(self) -> float:
        return safe_divide(self.reverse_complemented, self.n)

    def filtered_fraction(self, filter_name: str) -> float:
        return safe_divide(self.filtered.get(filter_name), self.n)


class ErrorRanges:
    """
    Representation of the lengths up to which a number of errors is allowed
    for partial adapter matches.

    >>> ErrorRanges(length=8, error_rate=0.1).lengths()
    [8]
    >>> ErrorRanges(length=19, error_rate=0.1).lengths()
    [9, 19]
    >>> ErrorRanges(length=20, error_rate=0.1).lengths()
    [9, 19, 20]
    >>> ErrorRanges(length=21, error_rate=0.1).lengths()
    [9, 19, 21]

    The entry at index i in the returned list is the length up to which
    i errors are allowed. For example, the list [9, 19, 23] describes that
    - 0 errors are allowed up to length 9
    - 1 error is allowed up to length 19
    - 2 errors are allowed up to length 23

    The last number in the list is always the length of the adapter sequence.
    """

    def __init__(self, length: int, error_rate: float):
        self.length = length
        self.error_rate = error_rate
        self._lengths = self._compute_lengths()

    def _compute_lengths(self) -> List[int]:
        lengths = [
            int(errors / self.error_rate) - 1
            for errors in range(1, int(self.error_rate * self.length) + 1)
        ]
        if not lengths or lengths[-1] < self.length:
            lengths.append(self.length)
        return lengths

    def __repr__(self):
        return (
            "ErrorRanges("
            f"length={self.length}, error_rate={self.error_rate}, _lengths={self._lengths})"
        )

    def __str__(self):
        """
        >>> str(ErrorRanges(length=8, error_rate=0.1))
        '1-8 bp: 0'
        >>> str(ErrorRanges(length=20, error_rate=0.1))
        '1-9 bp: 0; 10-19 bp: 1; 20 bp: 2'
        >>> str(ErrorRanges(length=23, error_rate=0.1))
        '1-9 bp: 0; 10-19 bp: 1; 20-23 bp: 2'
        """
        prev = 1
        s = ""
        for errors, r in enumerate(self._lengths[:-1]):
            s += f"{prev}-{r} bp: {errors}; "
            prev = r + 1
        if prev == self._lengths[-1]:
            s += f"{prev} bp: {len(self._lengths) - 1}"
        else:
            s += f"{prev}-{self._lengths[-1]} bp: {len(self._lengths) - 1}"
        return s

    def lengths(self):
        return self._lengths


def error_ranges(end_statistics: EndStatistics) -> str:
    length = end_statistics.effective_length
    error_rate = end_statistics.max_error_rate
    if end_statistics.allows_partial_matches:
        s = "\n" + str(ErrorRanges(length, error_rate))
    else:
        s = f" {int(error_rate * length)}"
    return "No. of allowed errors:" + s + "\n"


def histogram(end_statistics: EndStatistics, n: int, gc_content: float) -> str:
    """
    Return a formatted histogram. Include the no. of reads expected to be
    trimmed by chance (assuming a uniform distribution of nucleotides in the reads).

    adapter_statistics -- EndStatistics object
    adapter_length -- adapter length
    n -- total no. of reads.
    """
    sio = StringIO()

    print("length", "count", "expect", "max.err", "error counts", sep="\t", file=sio)
    for row in histogram_rows(end_statistics, n, gc_content):
        print(
            row.length,
            row.count,
            f"{row.expect:.1F}",
            row.max_err,
            " ".join(str(e) for e in row.error_counts),
            sep="\t",
            file=sio,
        )
    return sio.getvalue() + "\n"


@dataclass
class HistogramRow:
    """One row in the "trimmed lengths" histogram"""

    length: int
    count: int
    expect: float
    max_err: int
    error_counts: List[int]


def histogram_rows(
    end_statistics: EndStatistics,
    n: int,
    gc_content: float,
) -> Iterator[HistogramRow]:
    """
    Yield histogram rows

    Include the no. of reads expected to be
    trimmed by chance (assuming a uniform distribution of nucleotides in the reads).

    n -- total no. of reads.
    """
    d = end_statistics.lengths
    errors = end_statistics.errors

    match_probabilities = end_statistics.random_match_probabilities(
        gc_content=gc_content
    )
    for length in sorted(d):
        # when length surpasses adapter_length, the
        # probability does not increase anymore
        expect = n * match_probabilities[min(len(end_statistics.sequence), length)]
        count = d[length]
        max_errors = max(errors[length].keys())
        error_counts = [errors[length][e] for e in range(max_errors + 1)]
        row = HistogramRow(
            length=length,
            count=count,
            expect=expect,
            max_err=int(
                end_statistics.max_error_rate
                * min(length, end_statistics.effective_length)
            ),
            error_counts=error_counts,
        )
        yield row


class AdjacentBaseStatistics:
    def __init__(self, bases: Dict[str, int]):
        """ """
        self.bases: Dict[str, int] = bases
        self._warnbase: Optional[str] = None
        total = sum(self.bases.values())
        if total == 0:
            self._fractions = None
        else:
            self._fractions = []
            for base in ["A", "C", "G", "T", ""]:
                text = base if base != "" else "none/other"
                fraction = 1.0 * self.bases[base] / total
                self._fractions.append((text, 1.0 * self.bases[base] / total))
                if fraction > 0.8 and base != "":
                    self._warnbase = text
            if total < 20:
                self._warnbase = None

    def __repr__(self):
        return f"AdjacentBaseStatistics(bases={self.bases})"

    @property
    def should_warn(self) -> bool:
        return self._warnbase is not None

    @property
    def warnbase(self) -> Optional[str]:
        return self._warnbase

    def __str__(self) -> str:
        if not self._fractions:
            return ""
        sio = StringIO()
        print("Bases preceding removed adapters:", file=sio)
        for text, fraction in self._fractions:
            print(f"  {text}: {fraction:.1%}", file=sio)
        if self.should_warn:
            print("WARNING:", file=sio)
            print(
                f"    The adapter is preceded by '{self._warnbase}' extremely often.",
                file=sio,
            )
            print(
                "    The provided adapter sequence could be incomplete at its 5' end.",
                file=sio,
            )
            print("    Ignore this warning when trimming primers.", file=sio)
        return sio.getvalue()

    def as_json(self) -> Optional[Dict[str, int]]:
        if self._fractions:
            return {b: self.bases.get(b, 0) for b in ["A", "C", "G", "T", ""]}
        else:
            return None


def full_report(stats: Statistics, time: float, gc_content: float) -> str:  # noqa: C901
    """Print report to standard output."""
    if stats.n == 0:
        return "No reads processed!"
    if time == 0:
        time = 1e-6
    sio = StringIO()

    def print_s(*args, **kwargs):
        kwargs["file"] = sio
        print(*args, **kwargs)

    if sys.version_info[:2] <= (3, 6):
        micro = "u"
    else:
        micro = "µ"
    print_s(
        "Finished in {:.3F} s ({:.3F} {}s/read; {:.2F} M reads/minute).".format(
            time, 1e6 * time / stats.n, micro, stats.n / time * 60 / 1e6
        )
    )

    report = "\n=== Summary ===\n\n"
    if stats.paired:
        report += f"Total read pairs processed:      {stats.n:13,d}\n"
        for i in (0, 1):
            if stats.with_adapters[i] is not None:
                report += (
                    f"  Read {i+1} with adapter:           "
                    f"{stats.with_adapters[i]:13,d} ({stats.with_adapters_fraction[i]:.1%})\n"
                )
    else:
        report += f"Total reads processed:           {stats.n:13,d}\n"
        if stats.with_adapters[0] is not None:
            report += (
                f"Reads with adapters:             "
                f"{stats.with_adapters[0]:13,d} ({stats.with_adapters_fraction[0]:.1%})\n"
            )

    if stats.reverse_complemented is not None:
        report += (
            "Reverse-complemented:            "
            "{o.reverse_complemented:13,d} ({o.reverse_complemented_fraction:.1%})\n"
        )

    filter_report = format_filter_report(stats)
    if filter_report:
        report += "\n== Read fate breakdown ==\n"
        report += filter_report

    report += textwrap.dedent(
        """\
    {pairs_or_reads} written (passing filters): {o.written:13,d} ({o.written_fraction:.1%})

    Total basepairs processed: {o.total:13,d} bp
    """
    )
    if stats.paired:
        report += "  Read 1: {o.total_bp[0]:13,d} bp\n"
        report += "  Read 2: {o.total_bp[1]:13,d} bp\n"

    if stats.quality_trimmed is not None:
        report += (
            "Quality-trimmed:           "
            f"{stats.quality_trimmed:13,d} bp ({stats.quality_trimmed_fraction:.1%})\n"
        )
        if stats.paired:
            for i in (0, 1):
                if stats.quality_trimmed_bp[i] is not None:
                    report += f"  Read {i + 1}: {stats.quality_trimmed_bp[i]:13,d} bp\n"

    report += (
        "Total written (filtered):  "
        "{o.total_written_bp:13,d} bp ({o.total_written_bp_fraction:.1%})\n"
    )
    if stats.paired:
        report += "  Read 1: {o.written_bp[0]:13,d} bp\n"
        report += "  Read 2: {o.written_bp[1]:13,d} bp\n"
    pairs_or_reads = "Pairs" if stats.paired else "Reads"
    report = report.format(o=stats, pairs_or_reads=pairs_or_reads)
    print_s(report)

    warning = False
    for which_in_pair in (0, 1):
        for adapter_statistics in stats.adapter_stats[which_in_pair]:
            end_statistics = adapter_statistics.end_statistics()
            if end_statistics[0] is not None:
                total_front = sum(end_statistics[0].lengths.values())
            else:
                total_front = 0
            if end_statistics[1] is not None:
                total_back = sum(end_statistics[1].lengths.values())
            else:
                total_back = 0
            total = total_front + total_back
            reverse_complemented = adapter_statistics.reverse_complemented
            adapter = adapter_statistics.adapter
            if isinstance(adapter, BackAdapter):
                assert total_front == 0
            if isinstance(adapter, FrontAdapter):
                assert total_back == 0

            if stats.paired:
                extra = "First read: " if which_in_pair == 0 else "Second read: "
            else:
                extra = ""

            print_s("=" * 3, extra + "Adapter", adapter_statistics.name, "=" * 3)
            print_s()

            if isinstance(adapter_statistics, LinkedAdapterStatistics):
                print_s(
                    "Sequence: {}...{}; Type: linked; Length: {}+{}; "
                    "5' trimmed: {} times; 3' trimmed: {} times".format(
                        adapter_statistics.front.sequence,
                        adapter_statistics.back.sequence,
                        len(adapter_statistics.front.sequence),
                        len(adapter_statistics.back.sequence),
                        total_front,
                        total_back,
                    ),
                    end="",
                )
            else:
                assert isinstance(adapter, (SingleAdapter, AnywhereAdapter))
                print_s(
                    "Sequence: {}; Type: {}; Length: {}; Trimmed: {} times".format(
                        adapter.sequence,
                        adapter.description,
                        len(adapter.sequence),
                        total,
                    ),
                    end="",
                )
            if stats.reverse_complemented is not None:
                print_s(f"; Reverse-complemented: {reverse_complemented} times")
            else:
                print_s()
            if total == 0:
                print_s()
                continue
            if isinstance(adapter_statistics, AnywhereAdapterStatistics):
                assert isinstance(adapter, AnywhereAdapter)
                print_s(total_front, "times, it overlapped the 5' end of a read")
                print_s(
                    total_back, "times, it overlapped the 3' end or was within the read"
                )
                print_s()
                print_s("Minimum overlap:", adapter.min_overlap)
                print_s(error_ranges(adapter_statistics.front))
                print_s("Overview of removed sequences (5')")
                print_s(histogram(adapter_statistics.front, stats.n, gc_content))
                print_s()
                print_s("Overview of removed sequences (3' or within)")
                print_s(histogram(adapter_statistics.back, stats.n, gc_content))
            elif isinstance(adapter_statistics, LinkedAdapterStatistics):
                assert isinstance(adapter, LinkedAdapter)
                print_s()
                print_s(
                    f"Minimum overlap: "
                    f"{adapter.front_adapter.min_overlap}+{adapter.back_adapter.min_overlap}"
                )
                print_s(error_ranges(adapter_statistics.front))
                print_s(error_ranges(adapter_statistics.back))
                print_s("Overview of removed sequences at 5' end")
                print_s(histogram(adapter_statistics.front, stats.n, gc_content))
                print_s()
                print_s("Overview of removed sequences at 3' end")
                print_s(histogram(adapter_statistics.back, stats.n, gc_content))
            elif isinstance(adapter_statistics, FrontAdapterStatistics):
                assert isinstance(adapter, FrontAdapter)
                print_s()
                if adapter.allows_partial_matches:
                    print_s("Minimum overlap:", adapter.min_overlap)
                print_s(error_ranges(adapter_statistics.end))
                print_s("Overview of removed sequences")
                print_s(histogram(adapter_statistics.end, stats.n, gc_content))
            else:
                assert isinstance(adapter_statistics, BackAdapterStatistics)
                assert isinstance(adapter, BackAdapter)
                print_s()
                if adapter.allows_partial_matches:
                    print_s("Minimum overlap:", adapter.min_overlap)
                print_s(error_ranges(adapter_statistics.end))
                base_stats = AdjacentBaseStatistics(
                    adapter_statistics.end.adjacent_bases
                )
                warning = warning or base_stats.should_warn
                print_s(base_stats)
                print_s("Overview of removed sequences")
                print_s(histogram(adapter_statistics.end, stats.n, gc_content))

    if warning:
        print_s("WARNING:")
        print_s("    One or more of your adapter sequences may be incomplete.")
        print_s("    Please see the detailed output above.")

    return sio.getvalue().rstrip()


def format_filter_report(stats):
    report = ""
    for name, description in FILTERS.items():
        if name not in stats.filtered:
            continue
        value = stats.filtered[name]
        fraction = stats.filtered_fraction(name)
        line = (
            "{pairs_or_reads} "
            + (description + ":").ljust(27)
            + f"{value:13,d} ({fraction:.1%})\n"
        )
        report += line
    return report


def minimal_report(stats: Statistics, time: float, gc_content: float) -> str:
    """Create a minimal tabular report suitable for concatenation"""
    _ = time
    _ = gc_content

    fields = [
        "OK",
        stats.n,  # reads/pairs in
        stats.total,  # bases in
        stats.filtered.get("too_short", 0),  # reads/pairs
        stats.filtered.get("too_long", 0),  # reads/pairs
        stats.filtered.get("too_many_n", 0),  # reads/pairs
        stats.read_length_statistics.written_reads(),  # reads/pairs out
        stats.with_adapters[0] if stats.with_adapters[0] is not None else 0,  # reads
        stats.quality_trimmed_bp[0]
        if stats.quality_trimmed_bp[0] is not None
        else 0,  # bases
        stats.read_length_statistics.written_bp()[0],  # bases out
    ]
    if stats.paired:
        fields += [
            stats.with_adapters[1]
            if stats.with_adapters[1] is not None
            else 0,  # reads/pairs
            stats.quality_trimmed_bp[1]
            if stats.quality_trimmed_bp[1] is not None
            else 0,  # bases
            stats.read_length_statistics.written_bp()[1],  # bases
        ]

    warning = False
    for which_in_pair in (0, 1):
        for adapter_statistics in stats.adapter_stats[which_in_pair]:
            if isinstance(adapter_statistics, BackAdapterStatistics):
                if AdjacentBaseStatistics(
                    adapter_statistics.end.adjacent_bases
                ).should_warn:
                    warning = True
                    break
    if warning:
        fields[0] = "WARN"
    header = [
        "status",
        "in_reads",
        "in_bp",
        "too_short",
        "too_long",
        "too_many_n",
        "out_reads",
        "w/adapters",
        "qualtrim_bp",
        "out_bp",
    ]
    if stats.paired:
        header += ["w/adapters2", "qualtrim2_bp", "out2_bp"]
    return "\t".join(header) + "\n" + "\t".join(str(x) for x in fields)



import pytest

from utils import assert_files_equal, FilesDifferent, binomial


def test_files_different():
    with pytest.raises(FilesDifferent):
        assert_files_equal("simple.fasta", "simple.fastq")


def test_binomial():
    assert binomial(0, 0) == 1
    assert binomial(0, 1) == 0
    assert binomial(0, -1) == 0
    assert binomial(1, 0) == 1
    assert binomial(1, 1) == 1
    assert binomial(1, 2) == 0
    assert binomial(10, 5) == 10 * 9 * 8 * 7 * 6 // (2 * 3 * 4 * 5)
